'''Implements algorithms for clustering scans by
spectral similarity.
'''
import operator
import functools
import bisect
import json

from collections import deque

from scipy.special import comb

from ms_deisotope.task.log_utils import LogUtilsMixin
from ms_deisotope.data_source.dispatch import (
    SubsequenceMappingProxy as _SubsequenceMappingProxy,
    DynamicallyLoadingProxyResolver as _DynamicallyLoadingResolver)

from .similarity_methods import peak_set_similarity


peak_set_getter = operator.attrgetter("peak_set")
deconvoluted_peak_set_getter = operator.attrgetter("deconvoluted_peak_set")


def _ppm_error(x, y):
    return (x - y) / y


@functools.total_ordering
class SpectrumCluster(object):
    '''A collection of similar spectra.

    A :class:`SpectrumCluster` is compare-able by :attr:`neutral_mass`
    and acts as a :class:`collections.Sequence` over its members stored
    in :attr:`scans`.

    Attributes
    ----------
    scans: :class:`list`
        A list of spectra which are similar to each other.
    neutral_mass: :class:`float`
        The neutral mass of the representative spectrum's precursor.
    '''
    def __init__(self, scans=None, neutral_mass=None, average_similarity=None, annotations=None):
        if scans is None:
            scans = []
        if annotations is None:
            annotations = dict()
        self.scans = scans
        self.neutral_mass = None
        if neutral_mass:
            self.neutral_mass = neutral_mass
        elif self.scans:
            self.neutral_mass = self.scans[0].precursor_information.neutral_mass
        else:
            self.neutral_mass = 0.0
        if average_similarity is None and len(self) == 1:
            average_similarity = 1.0
        self._average_similarity = average_similarity
        self.annotations = annotations

    def __lt__(self, other):
        return self.neutral_mass < other.neutral_mass

    def __gt__(self, other):
        return self.neutral_mass > other.neutral_mass

    def __eq__(self, other):
        return (abs(self.neutral_mass - other.neutral_mass) / other.neutral_mass) < 1e-6

    def __repr__(self):
        return "SpectrumCluster(%f, %d)" % (self.neutral_mass, len(self))

    def __iter__(self):
        return iter(self.scans)

    def __len__(self):
        return len(self.scans)

    def __getitem__(self, i):
        return self.scans[i]

    def _invalidate(self):
        self._average_similarity = None

    def append(self, item, incremental_similarity=False):
        '''Add a new spectrum to the cluster.

        Parameters
        ----------
        item: :class:`~.ScanBase`
        '''
        if not self.neutral_mass:
            self.neutral_mass = item.precursor_information.neutral_mass
        if incremental_similarity:
            self._incremental_similarity(item)
        else:
            self._invalidate()
        self.scans.append(item)

    def _calculate_similarity_with(self, scan, from_ix=0, to_ix=None, *args, **kwargs):
        if from_ix is None:
            from_ix = 0
        if to_ix is None:
            to_ix = len(self)
        acc = []
        for member in self[from_ix:to_ix]:
            acc.append(peak_set_similarity(scan, member, *args, **kwargs))
        return acc

    def _incremental_similarity(self, scan, *args, **kwargs):
        new_sims = self._calculate_similarity_with(scan, *args, **kwargs)
        aggregate_size = comb(len(self), 2)
        n = (aggregate_size + len(new_sims))
        if n == 0:
            n = 1
        self._average_similarity = (aggregate_size * self.average_similarity() + sum(new_sims)
                                   ) / n

    def _full_similarity(self, *args, **kwargs):
        ratings = []
        n = len(self)
        for i in range(n):
            scan_i = self[i]
            for j in range(i + 1, n):
                scan_j = self[j]
                ratings.append(peak_set_similarity(scan_i, scan_j, *args, **kwargs))
        self._average_similarity = sum(ratings) / len(ratings)

    def average_similarity(self, *args, **kwargs):
        '''Calculate the within-cluster similarity among all cluster members
        and returns the average.

        All arguments are forwarded to :func:`~.peak_set_similarity`.

        If the cluster is a singleton or smaller, by definition its similarity
        is ``1.0``.

        Returns
        -------
        :class:`float`
        '''
        n = len(self)
        if n <= 1:
            return 1.0
        if self._average_similarity is not None:
            return self._average_similarity
        self._full_similarity()
        return self._average_similarity

    def to_dict(self):
        '''Convert the cluster to a JSON-safe :class:`dict`

        Returns
        -------
        :class:`dict`
        '''
        d = {}
        d['neutral_mass'] = self.neutral_mass
        d['size'] = len(self)
        d['average_similarity'] = self.average_similarity()
        scans = []
        for scan in self:
            scan_source = scan.source
            if scan_source is not None:
                source_name = scan_source.source_file
                if not isinstance(source_name, basestring):
                    if hasattr(source_name, 'name'):
                        source_name = source_name.name
            else:
                source_name = ":detatched:"
            scans.append({
                "id": scan.id,
                "source": source_name,
                "neutral_mass": scan.precursor_information.neutral_mass,
            })
        d['scans'] = scans
        return d


class SpectrumClusterCollection(object):
    '''A sorted :class:`~.Sequence` of :class:`SpectrumCluster` instances
    that supports searching by precursor mass.
    '''
    def __init__(self, clusters=None):
        if clusters is None:
            clusters = []
        self.clusters = list(clusters)

    def add(self, cluster):
        '''Add a new :class:`SpectrumCluster` to the collection,
        preserving sorted order.

        Parameters
        ----------
        cluster: :class:`SpectrumCluster`
            The cluster to add
        '''
        bisect.insort(self.clusters, cluster)

    def __getitem__(self, i):
        return self.clusters[i]

    def __setitem__(self, i, value):
        self.clusters[i] = value

    def __len__(self):
        return len(self.clusters)

    def __iter__(self):
        return iter(self.clusters)

    def __repr__(self):
        template = "{self.__class__.__name__}({size})"
        size = len(self)
        return template.format(self=self, size=size)

    def _binary_search(self, mass, error_tolerance=1e-5):
        array = self.clusters
        n = len(array)
        lo = 0
        hi = n

        while hi != lo:
            mid = (hi + lo) // 2
            y = array[mid].neutral_mass
            err = (y - mass) / mass
            if hi - lo == 1:
                best_index = mid
                best_error = abs(err)
                i = mid - 1
                while i >= 0:
                    x = array[i]
                    err = abs((x.neutral_mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1
                lo_index = i + 1
                i = mid + 1
                while i < n:
                    x = array[i]
                    err = abs((x.neutral_mass - mass) / mass)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                hi_index = i
                return best_index, lo_index, hi_index
            elif err > 0:
                hi = mid
            else:
                lo = mid
        return 0, 0, 0

    def find(self, mass, error_tolerance=1e-5):
        '''Finds the cluster whose precursor mass is closest to
        ``mass`` within ``error_tolerance`` ppm error.

        Parameters
        ----------
        mass: :class:`float`
            The mass to search for.
        error_tolerance: :class:`float`
            The PPM error tolerance to use. Defaults to 1e-5.

        Returns
        -------
        :class:`SpectrumCluster` or :const:`None`
        '''
        target_ix, _, _ = self._binary_search(mass, error_tolerance)
        target = self[target_ix]
        if abs(target.neutral_mass - mass) / mass > error_tolerance:
            return None
        return target

    def find_all(self, mass, error_tolerance=1e-5):
        '''Finds all clusters whose precursor mass is within ``error_tolerance``
        ppm of ``mass``.

        Parameters
        ----------
        mass: :class:`float`
            The mass to search for.
        error_tolerance: :class:`float`
            The PPM error tolerance to use. Defaults to 1e-5.

        Returns
        -------
        :class:`list`
        '''
        _, lo_ix, hi_ix = self._binary_search(mass, error_tolerance)
        result = [
            target for target in self[lo_ix:hi_ix]
            if abs(target.neutral_mass - mass) / mass <= error_tolerance
        ]
        return result


class _PeakGetterStrategyBase(object):
    def __call__(self, scan):
        return self.peaks(scan)


class _FittedPeakAccessorStrategy(_PeakGetterStrategyBase):
    def peaks(self, scan):
        return scan.peak_set

    def tic(self, scan):
        return scan.tic.centroided()


class _DeconvolutedPeakAccessorStrategy(_PeakGetterStrategyBase):
    def peaks(self, scan):
        return scan.deconvoluted_peak_set

    def tic(self, scan):
        return scan.tic.deconvoluted()


class _DynamicPeakAccessorStrategy(_PeakGetterStrategyBase):

    def _ensure_peak_set(self, scan):
        if scan.peak_set is None:
            scan.pick_peaks()
        return scan.peak_set

    def peaks(self, scan):
        if scan.deconvoluted_peak_set is not None:
            return scan.deconvoluted_peak_set
        else:
            return self._ensure_peak_set(scan)

    def tic(self, scan):
        if scan.deconvoluted_peak_set is not None:
            return scan.tic.deconvoluted()
        else:
            self._ensure_peak_set(scan)
            return scan.tic.centroided()


class ScanClusterBuilder(LogUtilsMixin):
    """Clusters spectra based upon peak pattern similarity

    Attributes
    ----------
    clusters : :class:`SpectrumClusterCollection`
        The clusters built so far
    minimum_similarity : float
        The minimum similarity score needed to consider two spectra
        similar enough to form a cluster
    peak_getter : :class:`Callable`
        A function to call on each spectrum to retrieve the peak list
        to cluster over
    precursor_error_tolerance : float
        The maximum precursor mass error (in PPM) to permit between
        two spectra to consider comparing them
    track_incremental_similarity : bool
        Whether to incrementally update a cluster's similarity when it
        grows.
    """

    @classmethod
    def _guess_peak_getter(cls, getter):
        if getter is None:
            return _DynamicPeakAccessorStrategy()
        if callable(getter):
            return getter
        if isinstance(getter, basestring):
            if getter == "d":
                return _DeconvolutedPeakAccessorStrategy()
            if getter in ('p', 'c'):
                return _FittedPeakAccessorStrategy()
            else:
                raise ValueError("Cannot infer peak getter strategy from %r" % (getter, ))

        raise ValueError(
            "Cannot infer peak set getter strategy from %r" % (getter, ))

    def __init__(self, clusters=None, precursor_error_tolerance=1e-5, minimum_similarity=0.1,
                 peak_getter=None, track_incremental_similarity=True):
        peak_getter = self._guess_peak_getter(peak_getter)
        if clusters is None:
            clusters = []
        self.clusters = SpectrumClusterCollection(clusters)
        self.precursor_error_tolerance = precursor_error_tolerance
        self.minimum_similarity = minimum_similarity
        self.peak_getter = peak_getter
        self.track_incremental_similarity = track_incremental_similarity

    def _binsearch_simple(self, x):
        n = len(self)
        lo = 0
        hi = n

        while hi != lo:
            mid = (hi + lo) // 2
            y = self[mid].neutral_mass
            err = y - x
            # Do refinement in `find_best_cluster_for_scan`
            if hi - lo == 1:
                return mid
            elif err > 0:
                hi = mid
            else:
                lo = mid
        return 0

    def peak_set_similarity(self, scan_i, scan_j):
        '''Calculate the similarity between the peaks of
        ``scan_i`` and ``scan_j``, where the peaks are
        those points given by :attr:`peak_getter`

        Parameters
        ----------
        scan_i: :class`~.ScanBase`
        scan_j: :class`~.ScanBase`

        Returns
        -------
        :class:`float`
        '''
        peak_set_a = self.peak_getter(scan_i)
        peak_set_b = self.peak_getter(scan_j)
        return peak_set_similarity(
            peak_set_a, peak_set_b)

    def find_best_cluster_for_scan(self, scan):
        '''Locate the best cluster to add ``scan`` to according to
        precursor mass and peak set similarity.

        Parameters
        ----------
        scan: :class:`~.ScanBase`

        Returns
        -------
        :class:`SpectrumCluster`
        '''
        best_cluster = None
        best_similarity = 0.0
        n = len(self.clusters)
        if n == 0:
            return best_cluster

        center_i = self._binsearch_simple(scan.precursor_information.neutral_mass)
        i = center_i

        while i >= 0:
            cluster = self.clusters[i]
            if abs(_ppm_error(scan.precursor_information.neutral_mass,
                              cluster.neutral_mass)) > self.precursor_error_tolerance:
                break
            similarity = self.peak_set_similarity(scan, cluster[0])
            i -= 1
            if similarity > best_similarity and similarity > self.minimum_similarity:
                best_similarity = similarity
                best_cluster = cluster

        i = center_i + 1
        while i < n:
            cluster = self.clusters[i]
            if abs(_ppm_error(scan.precursor_information.neutral_mass,
                              cluster.neutral_mass)) > self.precursor_error_tolerance:
                break
            similarity = self.peak_set_similarity(scan, cluster[0])
            i += 1
            if similarity > best_similarity and similarity > self.minimum_similarity:
                best_similarity = similarity
                best_cluster = cluster
        return best_cluster

    def add_scan(self, scan):
        '''Add ``scan`` to the cluster collection, adding it to the best
        matching cluster, or starting a new cluster around it if no good
        match can be found.

        Parameters
        ----------
        scan: :class:`~.ScanBase`
        '''
        best_cluster = self.find_best_cluster_for_scan(scan)
        if best_cluster:
            best_cluster.append(scan, incremental_similarity=self.track_incremental_similarity)
        else:
            self.clusters.add(SpectrumCluster([scan]))

    def __iter__(self):
        return iter(self.clusters)

    def __len__(self):
        return len(self.clusters)

    def __getitem__(self, i):
        return self.clusters[i]

    def _get_tic(self, scan):
        try:
            return self.peak_getter.tic(scan)
        except AttributeError:
            return sum(p.intensity for p in self.peak_getter(scan))

    @classmethod
    def cluster_scans(cls, scans, precursor_error_tolerance=1e-5, minimum_similarity=0.1,
                      peak_getter=None, sort=True, track_incremental_similarity=False):
        '''Cluster scans by precursor mass and peak set similarity.

        Parameters
        ----------
        scans: :class:`Iterable`
            An iterable of :class:`Scan`-like objects
        precursor_error_tolerance: :class:`float`
            The PPM mass accuracy threshold for precursor mass differences to
            tolerate when deciding whether to compare spectra. Defaults to 1e-5.
        minimum_similarity: :class:`float`
            The minimum peak set similarity required to consider adding a spectrum
            to a cluster. Defaults to 0.1
        peak_getter: :class:`Callable`
            A callable object used to get peaks from elements of ``scans``.
        sort: :class:`bool`
            Whether or not to sort spectra by their total ion current before clustering.
        '''
        self = cls([], precursor_error_tolerance, minimum_similarity,
                   peak_getter, track_incremental_similarity)
        if sort:
            if len(scans) > 100:
                self.log("Sorting Scans By TIC")
            scans = sorted(scans, key=self._get_tic, reverse=True)
        if len(scans) > 100:
            self.log("Clustering %d Scans" % (len(scans), ))
        n = len(scans)
        report_interval = max(min(n // 10, 1000), 50)
        for i, scan in enumerate(scans):
            if i % report_interval == 0 and i:
                self.log("... Handled %d Scans (%0.2f%%)" % (i, i * 100.0 / n))
            self.add_scan(scan)
        return self.clusters

    @classmethod
    def iterative_clustering(cls, scans, precursor_error_tolerance=1e-5, similarity_thresholds=None,
                             peak_getter=None):
        '''Cluster scans by precursor mass and peak set similarity, iteratively refining
        clusters with increasing similarity threshold requirements.

        Parameters
        ----------
        scans: :class:`Iterable`
            An iterable of :class:`Scan`-like objects
        precursor_error_tolerance: :class:`float`
            The PPM mass accuracy threshold for precursor mass differences to
            tolerate when deciding whether to compare spectra. Defaults to 1e-5.
        similarity_thresholds: :class:`Sequence` of :class:`float`
            A series of similarity thresholds to apply as spectra are added to clusters
            and as clusters are iteratively refined.
        peak_getter: :class:`Callable`
            A callable object used to get peaks from elements of ``scans``.
        '''
        peak_getter = cls._guess_peak_getter(peak_getter)
        if similarity_thresholds is None:
            similarity_thresholds = [0.1, .4, 0.7]
        singletons = []
        to_bisect = [scans]
        logger = LogUtilsMixin()
        track_similarity = [False] * (len(similarity_thresholds) - 1)
        track_similarity.append(True)
        for similarity_threshold, track in zip(similarity_thresholds, track_similarity):
            logger.log("Clustering with Threshold %0.2f" % (similarity_threshold, ))
            next_to_bisect = []
            if len(to_bisect) > 1:
                logger.log("Refining %d Clusters" % (len(to_bisect)))
            elif to_bisect:
                logger.log("Clustering %d Scans" % (len(to_bisect[0])))
            else:
                logger.log("Nothing to cluster...")
                return SpectrumClusterCollection([])
            n = len(to_bisect)
            report_interval = max(min(n // 10, 1000), 1)
            for i, group in enumerate(to_bisect):
                if i % report_interval == 0:
                    logger.log("... Handling Batch %d (%d Scans)" % (i, len(group)))
                clusters = cls.cluster_scans(
                    group, precursor_error_tolerance,
                    minimum_similarity=similarity_threshold,
                    peak_getter=peak_getter, sort=True, track_incremental_similarity=track)
                for cluster in clusters:
                    if len(cluster) == 1:
                        singletons.append(cluster)
                    else:
                        next_to_bisect.append(cluster)
            logger.log("%d Singletons and %d Groups" % (len(singletons), len(next_to_bisect)))
            to_bisect = next_to_bisect
        return SpectrumClusterCollection(sorted(list(singletons) + list(to_bisect)))


cluster_scans = ScanClusterBuilder.cluster_scans
iterative_clustering = ScanClusterBuilder.iterative_clustering


class ScanClusterWriterBase(object):
    '''A base class for writing :class:`ScanCluster` objects to an
    I/O stream like a file.

    Attributes
    ----------
    stream: :class:`io.IOBase`
        The stream to write the clusters to
    metadata: :class:`dict`
        A set of key-value pairs that describe this
        collection.
    '''

    def __init__(self, stream, metadata=None):
        self.stream = stream
        self.metadata = metadata or {}
        self._wrote_metadata = False

    def _write(self, data):
        self.stream.write(data)

    def save(self, cluster):
        '''Write ``cluster`` to the output stream, recording its
        members and calculating it's average similarity.

        Parameters
        ----------
        cluster: :class:`SpectrumCluster`
            The spectrum cluster to write out
        '''
        if not self._wrote_metadata:
            self.write_metadata()
            self._wrote_metadata = True
        self._save(cluster)

    def _save(self, cluster):
        raise NotImplementedError()

    def save_all(self, clusters):
        '''Write each :class:`SpectrumCluster` in ``clusters`` out,
        calling :meth:`save` on each one.

        Parameters
        ----------
        clusters: :class:`collections.Iterable` of :class:`SpectrumCluster`
            The spectrum clusters to write out
        '''
        raise NotImplementedError()

    def add_metadata(self, key, value):
        '''Add metadata to the writer. That metadata will be flushed
        out upon starting to write clusters out.

        Parameters
        ----------
        key: :class:`str`
            The metadata element's name
        value: :class:`str`, :class:`float`
            The metadata element's value
        '''
        if self.wrote_metadata:
            raise TypeError(
                "Cannot add additional metadata, the metadata has already been written")
        self.metadata[key] = value

    def write_metadata(self):
        '''Write the accumulated metadata out in a format-appropriate
        manner at the top of the file.
        '''
        if self._wrote_metadata:
            raise TypeError("Already wrote metadata!")


class ScanClusterWriter(ScanClusterWriterBase):
    '''Writes :class:`ScanCluster` objects to a hierarchical text stream

    Parameters
    ----------
    stream: :class:`io.IOBase`
        The stream to write the clusters to
    '''

    def __init__(self, stream, metadata=None):
        super(ScanClusterWriter, self).__init__(stream, metadata)

    def write_metadata(self):
        for key, value in self.metadata.items():
            self._write("#%s = %s\n" % (key, value))
        self._write("\n")

    def _save(self, cluster):
        '''Write ``cluster`` as a tab delimited tree, recording its
        members and calculating it's average similarity.

        Parameters
        ----------
        cluster: :class:`SpectrumCluster`
            The spectrum cluster to write out
        '''
        self._write("%f\t%d\t%f\n" % (cluster.neutral_mass, len(cluster), cluster.average_similarity()))
        for member in cluster:
            member_source = member.source
            if member_source is not None:
                source_name = member_source.source_file
                if not isinstance(source_name, basestring):
                    if hasattr(source_name, 'name'):
                        source_name = source_name.name
            else:
                source_name = ":detatched:"
            self._write("\t%s\t%s\n" % (source_name, member.id))
        self._write('\n')

    def save_all(self, clusters):
        '''Write each :class:`SpectrumCluster` in ``clusters`` out,
        calling :meth:`save` on each one.

        Parameters
        ----------
        clusters: :class:`collections.Iterable` of :class:`SpectrumCluster`
            The spectrum clusters to write out
        '''
        for cluster in clusters:
            self.save(cluster)


class JSONScanClusterWriter(ScanClusterWriterBase):
    '''A :class:`ScanClusterWriterBase` that uses JSON
    lines.
    '''

    def write_metadata(self):
        json.dump(self.metadata, self.stream)
        self._write("\n")

    def _save(self, cluster):
        json.dump(cluster.to_dict(), self.stream)
        self._write("\n")

    def save_all(self, clusters):
        for cluster in clusters:
            self.save(cluster)


class ScanClusterReaderBase(object):
    '''Base class for reading spectrum clusters from disk.

    Attributes
    ----------
    stream: :class:`io.IOBase`
        The stream to read the clusters from
    resolver_map: :class:`dict`
        A mapping from scan source name to a :class:`Callable`
        which will return a :class:`~.ScanBase` object representing
        that spectrum.
    metadata: :class:`dict`
        A set of key-value pairs that describe this
        collection.
    clusters: :class:`list`
        The read clusters.
    '''
    def __init__(self, stream, resolver_map):
        self.stream = stream
        self.resolver_map = resolver_map
        self.clusters = list()
        self.metadata = {}
        self._generator = None

    def _resolve(self, source, scan_id):
        resolver = self.resolver_map[source]
        return resolver(scan_id)

    def _parse(self):
        '''Parse the cluster collection from :attr:`stream`
        '''
        self._load_metadata()
        return self._load_clusters()

    def _load_metadata(self):
        '''Read the metadata header from :attr:`stream`.
        '''
        raise NotImplementedError()

    def _load_clusters(self):
        '''Read the data describing :class:`SpectrumCluster` objects from
        :attr:`stream`.
        '''
        raise NotImplementedError()

    def __iter__(self):
        return self

    def __next__(self):
        '''Advance the iterator, retrieving the next :class:`SpectrumCluster`

        Returns
        -------
        :class:`SpectrumCluster`
        '''
        if self._generator is None:
            self._generator = self._parse()
        return next(self._generator)

    def next(self):
        '''Advance the iterator, retrieving the next :class:`SpectrumCluster`

        Returns
        -------
        :class:`SpectrumCluster`
        '''
        return self.__next__()


class ScanClusterReader(ScanClusterReaderBase):
    '''Reads :class:`SpectrumCluster` objects from hierarchical text files written by
    :class:`ScanClusterWriter`.
    '''
    def __init__(self, stream, resolver_map):
        super(ScanClusterReader, self).__init__(stream, resolver_map)
        self._line_buffer = deque()

    def _next_line(self):
        if self._line_buffer:
            return self._line_buffer.popleft()
        return self.stream.readline()

    def _return_line(self, line):
        self._line_buffer.append(line)

    def _stream_lines(self):
        line = self._next_line()
        while line:
            yield line
            line = self._next_line()

    def _load_metadata(self):
        line = self._next_line()
        while line.startswith("#"):
            key, value = line.strip().split(" = ", 1)
            try:
                value = float(value)
            except ValueError:
                value = str(value)
            self.metadata[key] = value
        self._return_line(line)

    def _load_clusters(self):
        current_cluster = []
        mass = None
        similarity = None
        for line in self._stream_lines():
            line = line.rstrip()
            if not line or not line.startswith('\t'):
                if current_cluster:
                    yield SpectrumCluster(current_cluster, mass, similarity)
                    current_cluster = []
                    mass = None
                    similarity = None
                if line:
                    tokens = line.split('\t')
                    mass = float(tokens[0])
                    # size = int(tokens[1])
                    similarity = float(tokens[2])
            elif line.startswith("\t"):
                tokens = line.split("\t")
                source = tokens[1]
                scan_id = tokens[2]
                scan = self._resolve(source, scan_id)
                current_cluster.append(scan)
        if current_cluster:
            yield SpectrumCluster(current_cluster, mass, similarity)


class JSONScanClusterReader(ScanClusterReader):
    '''Reads :class:`SpectrumCluster` objects from hierarchical text files written by
    :class:`JSONScanClusterWriter`.
    '''
    def _load_metadata(self):
        return json.loads(self.stream.readline())

    def _load_clusters(self):
        line = self.stream.readline()
        while line:
            data = json.loads(line)
            mass = data['neutral_mass']
            similarity = data['average_similarity']
            scans = []
            for scan_bundle in data['scans']:
                scans.append(self._resolve(scan_bundle['source'], scan_bundle['id']))
            yield SpectrumCluster(scans, mass, similarity)
            line = self.stream.readline()



cimport cython
from cython.parallel cimport prange
from libc.math cimport floor, sqrt
from libc.stdlib cimport malloc, free
from cpython cimport PyErr_SetString

from cpython.list cimport PyList_Size, PyList_GET_ITEM
from cpython.tuple cimport PyTuple_GET_ITEM, PyTuple_Size

from ms_peak_picker._c.peak_index cimport PeakIndex

from ms_peak_picker._c.peak_set cimport PeakBase, FittedPeak, PeakSet
from ms_deisotope._c.peak_set cimport DeconvolutedPeak, DeconvolutedPeakSet

cimport numpy as cnp
import numpy as np

cdef object zeros = np.zeros


ctypedef fused peak_collection:
    PeakSet
    PeakIndex
    DeconvolutedPeakSet
    object


cpdef enum SimilarityMetrics:
    dot_product
    normalized_dot_product


@cython.boundscheck(False)
@cython.cdivision(True)
cpdef double peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0) except -1:
    """Computes the normalized dot product, also called cosine similarity between
    two peak sets, a similarity metric ranging between 0 (dissimilar) to 1.0 (similar).

    Parameters
    ----------
    peak_set_a : Iterable of Peak-like
    peak_set_b : Iterable of Peak-like
        The two peak collections to compare. It is usually only useful to
        compare the similarity of peaks of the same class, so the types
        of the elements of `peak_set_a` and `peak_set_b` should match.
    precision : int, optional
        The precision of rounding to use when binning spectra. Defaults to 0

    Returns
    -------
    float
        The similarity between peak_set_a and peak_set_b. Between 0.0 and 1.0
    """

    cdef:
        tuple binned_spectra
        double n_a, n_b, n_ab, z, a, b
        double hi, scaler
        PeakBase peak
        double *bin_a
        double *bin_b
        long i, n, k, index

    n_a = 0
    n_b = 0
    a = 0
    b = 0
    z = 0

    # Determine maximum value
    hi = 0
    if peak_collection is PeakIndex:
        peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
        hi = peak.mz
        peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)
        hi = max(peak.mz, hi)
    elif peak_collection is object:
        hi = 0
        for obj in peak_set_a:
            peak = <PeakBase>obj
            if peak.mz > hi:
                hi = peak.mz
        for obj in peak_set_b:
            peak = <PeakBase>obj
            if peak.mz > hi:
                hi = peak.mz
    elif peak_collection is PeakSet:
        peak = peak_set_a.getitem(peak_set_a.get_size() - 1)
        hi = peak.mz
        peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
        hi = max(peak.mz, hi)
    elif peak_collection is DeconvolutedPeakSet:
        peak = peak_set_a._mz_ordered[-1]
        hi = peak.mz
        peak = peak_set_b._mz_ordered[-1]
        hi = max(peak.mz, hi)

    # Calculate bin array size and allocate memory
    scaler = 10 ** precision
    k = <long>((hi * scaler) + 1)
    bin_a = <double*>malloc(k * sizeof(double))
    bin_b = <double*>malloc(k * sizeof(double))
    if bin_a == NULL or bin_b == NULL:
        if bin_a != NULL:
            free(bin_a)
        if bin_b != NULL:
            free(bin_b)
        PyErr_SetString(MemoryError, "Could not allocate bins")
        return -1
    for i in range(k):
        bin_a[i] = bin_b[i] = 0

    # Fill bins
    if peak_collection is object:
        n = len(peak_set_a)
    else:
        n = peak_set_a.get_size()
    for i in range(n):
        if peak_collection is PeakIndex:
            peak = peak_set_a.peaks.getitem(i)
        elif peak_collection is object:
            peak = peak_set_a[i]
        else:
            peak = peak_set_a.getitem(i)
        index = int(peak.mz * scaler)
        bin_a[index] += peak.intensity

    if peak_collection is object:
        n = len(peak_set_b)
    else:
        n = peak_set_b.get_size()
    for i in range(n):
        if peak_collection is PeakIndex:
            peak = peak_set_b.peaks.getitem(i)
        elif peak_collection is object:
            peak = peak_set_b[i]
        else:
            peak = peak_set_b.getitem(i)
        index = int(peak.mz * scaler)
        bin_b[index] += peak.intensity

    # Calculate dot product and normalizers in parallel
    n = k
    for i in prange(n, nogil=True, schedule='static', num_threads=4):
        a = bin_a[i]
        b = bin_b[i]
        n_a += a * a
        n_b += b * b
        z += a * b

    free(bin_a)
    free(bin_b)
    n_ab = sqrt(n_a) * sqrt(n_b)
    if n_ab == 0:
        return 0.0
    return z / n_ab


@cython.boundscheck(False)
cpdef tuple bin_peaks(peak_collection peak_set_a, peak_collection peak_set_b, int precision=0):
    cdef:
        double hi, scaler
        PeakBase peak
        cnp.ndarray[double, ndim=1] bin_a, bin_b
        size_t i, n, index

    if peak_collection is PeakIndex:
        hi = 0
        peak = peak_set_a.peaks.getitem(peak_set_a.get_size() - 1)
        hi = peak.mz
        peak = peak_set_b.peaks.getitem(peak_set_b.get_size() - 1)
        hi = max(peak.mz, hi)
    elif peak_collection is object:
        hi = 0
        for obj in peak_set_a:
            peak = <PeakBase>obj
            if peak.mz > hi:
                hi = peak.mz
        for obj in peak_set_b:
            peak = <PeakBase>obj
            if peak.mz > hi:
                hi = peak.mz
    elif peak_collection is DeconvolutedPeakSet:
        peak = peak_set_a._mz_ordered[-1]
        hi = peak.mz
        peak = peak_set_b._mz_ordered[-1]
        hi = max(peak.mz, hi)
    else:
        hi = 0
        peak = peak_set_a.getitem(peak_set_a.get_size() - 1)
        hi = peak.mz
        peak = peak_set_b.getitem(peak_set_b.get_size() - 1)
        hi = max(peak.mz, hi)
    scaler = 10 ** precision
    hi = (hi * scaler) + 1
    bin_a = zeros(<int>hi)
    bin_b = zeros(<int>hi)

    if peak_collection is object:
        n = len(peak_set_a)
    else:
        n = peak_set_a.get_size()
    for i in range(n):
        if peak_collection is PeakIndex:
            peak = peak_set_a.peaks.getitem(i)
        elif peak_collection is object:
            peak = peak_set_a[i]
        else:
            peak = peak_set_a.getitem(i)
        index = int(peak.mz * scaler)
        bin_a[index] += peak.intensity

    if peak_collection is object:
        n = len(peak_set_b)
    else:
        n = peak_set_b.get_size()
    for i in range(n):
        if peak_collection is PeakIndex:
            peak = peak_set_b.peaks.getitem(i)
        elif peak_collection is object:
            peak = peak_set_b[i]
        else:
            peak = peak_set_b.getitem(i)
        index = int(peak.mz * scaler)
        bin_b[index] += peak.intensity
    return (bin_a, bin_b)


@cython.boundscheck(False)
@cython.cdivision(True)
cdef double calculate_dot_product(cnp.ndarray[double, ndim=1] bin_a, cnp.ndarray[double, ndim=1] bin_b, bint normalize=True):
    cdef:
        double n_a, n_b, n_ab, z, a, b
        long i, n
    with nogil:
        n_a = 0
        n_b = 0
        z = 0
        n = bin_a.shape[0]
        for i in prange(n, schedule='static'):
            a = bin_a[i]
            b = bin_b[i]
            n_a += a * a
            n_b += b * b
            z += a * b
        if normalize:
            n_ab = sqrt(n_a) * sqrt(n_b)
            if n_ab == 0:
                return 0.0
            return z / n_ab
        return z


cdef list convolve_peak_sets(PeakSet peak_set_a, PeakSet peak_set_b, double error_tolerance=2e-5):
    cdef:
        size_t i, n, j, m
        FittedPeak peak
        FittedPeak other
        PeakSet peaks_slice
        list peak_pairs
        list pairs_for_peak

    peak_pairs = []
    n = peak_set_a.get_size()
    for i in range(n):
        pairs_for_peak = []
        peak = peak_set_a.getitem(i)
        peaks_slice = peak_set_b._between(
            peak.mz - peak.mz * error_tolerance, peak.mz + peak.mz * error_tolerance)
        m = peaks_slice.get_size()
        for j in range(m):
            other = peaks_slice.getitem(j)
            pairs_for_peak.append((peak, other))
        peak_pairs.append(pairs_for_peak)
    return peak_pairs


cdef double convolved_dot_product(list peak_pairs):
    cdef:
        size_t i, j, n, m
        PeakBase peak, other
        list pairs
        tuple pair
        double d

    d = 0.0
    n = PyList_Size(peak_pairs)
    for i in range(n):
        pairs = <list>PyList_GET_ITEM(peak_pairs, i)
        m = PyList_Size(pairs)
        for j in range(m):
            pair = <tuple>PyList_GET_ITEM(pairs, j)
            peak = <PeakBase>PyTuple_GET_ITEM(pair, 0)
            other = <PeakBase>PyTuple_GET_ITEM(pair, 1)
            d += peak.intensity * other.intensity
    return d


cdef double convolve_peak_sets2_fitted(PeakSet peak_set_a, PeakSet peak_set_b, double error_tolerance=2e-5):
    cdef:
        size_t i, n, j, m
        FittedPeak peak
        FittedPeak other
        PeakSet peaks_slice
        double acc

    acc = 0.0
    n = peak_set_a.get_size()
    for i in range(n):
        peak = peak_set_a.getitem(i)
        peaks_slice = peak_set_b._between(
            peak.mz - peak.mz * error_tolerance, peak.mz + peak.mz * error_tolerance)
        m = peaks_slice.get_size()
        for j in range(m):
            other = peaks_slice.getitem(j)
            acc += peak.intensity * other.intensity
    return acc


cdef double convolve_peak_sets2_deconvoluted(DeconvolutedPeakSet peak_set_a, DeconvolutedPeakSet peak_set_b, double error_tolerance=2e-5):
    cdef:
        size_t i, n, j, m
        DeconvolutedPeak peak
        DeconvolutedPeak other
        tuple peaks_slice
        double acc

    acc = 0.0
    n = peak_set_a.get_size()
    for i in range(n):
        peak = peak_set_a.getitem(i)
        peaks_slice = peak_set_b.all_peaks_for(peak.neutral_mass, error_tolerance)
        m = PyTuple_Size(peaks_slice)
        for j in range(m):
            other = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks_slice, j)
            acc += peak.intensity * other.intensity
    return acc


cpdef double ppm_peak_set_similarity(peak_collection peak_set_a, peak_collection peak_set_b, double error_tolerance=2e-5):
    cdef:
        double ab, aa, bb
    if peak_collection is PeakSet:
        ab = convolve_peak_sets2_fitted(peak_set_a, peak_set_b, error_tolerance)
        aa = convolve_peak_sets2_fitted(peak_set_a, peak_set_a, error_tolerance)
        bb = convolve_peak_sets2_fitted(peak_set_b, peak_set_b, error_tolerance)
    elif peak_collection is PeakIndex:
        ab = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_b.peaks, error_tolerance)
        aa = convolve_peak_sets2_fitted(peak_set_a.peaks, peak_set_a.peaks, error_tolerance)
        bb = convolve_peak_sets2_fitted(peak_set_b.peaks, peak_set_b.peaks, error_tolerance)
    elif peak_collection is DeconvolutedPeakSet:
        ab = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_b, error_tolerance)
        aa = convolve_peak_sets2_deconvoluted(peak_set_a, peak_set_a, error_tolerance)
        bb = convolve_peak_sets2_deconvoluted(peak_set_b, peak_set_b, error_tolerance)
    elif peak_collection is object:
        raise TypeError("Cannot handle objects of type %s" % type(peak_set_a))

    return (ab) / (sqrt((aa)) * sqrt((bb)))



try:
    from Queue import Empty as QueueEmpty
except ImportError:
    from queue import Empty as QueueEmpty

from ms_deisotope.data_source.common import ProcessedScan
from ms_deisotope.task import TaskBase, CallInterval

from .process import (
    SCAN_STATUS_SKIP, DONE)


class ScanCollator(TaskBase):
    """Collates incoming scan bunches from multiple
    ScanTransformingProcesses, passing them along in
    the correct order.

    Attributes
    ----------
    count_jobs_done : int
        The number of scan bunches taken from :attr:`queue`
    count_since_last : int
        The number of work-cycles since the last scan bunch
        has been yielded
    done_event : multiprocessing.Event
        An IPC Event to indicate that all scan ids have been
        sent to the worker processes
    helper_producers : list
        A list of ScanTransformingProcesses
    include_fitted : bool
        Whether or not to save the raw fitted peaks for each
        scan produced. When this is `False`, they will be
        discarded and memory will be saved
    last_index : int
        The index of the last scan yielded through the iterator
        loop. This controls the next scan to be yielded and any
        waiting conditions
    primary_worker : ScanTransformingProcess
        The first worker to start consuming scans which will dictate
        the first handled index. Is required to run in isolation
        from other worker processes to insure that the first scan
        arrives in order
    queue : multiprocessing.Queue
        The IPC queue that all workers place their results on
        to be consumed and yielded in order
    started_helpers : bool
        Whether or not the additional workers in :attr:`helper_producers`
        have been started or not
    waiting : dict
        A mapping from scan index to `Scan` object. Used to serve
        scans through the iterator when their index is called for
    """
    _log_received_scans = False

    def __init__(self, queue, done_event, helper_producers=None, primary_worker=None,
                 include_fitted=False, input_queue=None):
        if helper_producers is None:
            helper_producers = []
        self.queue = queue
        self.last_index = None
        self.count_jobs_done = 0
        self.count_since_last = 0
        self.waiting = {}
        self.done_event = done_event
        self.helper_producers = helper_producers
        self.started_helpers = False
        self.primary_worker = primary_worker
        self.include_fitted = include_fitted
        self.input_queue = input_queue

    def all_workers_done(self):
        if self.done_event.is_set():
            if self.primary_worker.all_work_done():
                for helper in self.helper_producers:
                    if not helper.all_work_done():
                        return False
                return True
            else:
                return False
        return False

    def store_item(self, item, index):
        """Stores an incoming work-item for easy
        access by its `index` value. If configuration
        requires it, this will also reduce the number
        of peaks in `item`.

        Parameters
        ----------
        item : str or ProcessedScan
            Either a stub indicating why this work item
            is not
        index : int
            Scan index to store
        """
        if self._log_received_scans:
            self.log("-- received %d: %s" % (index, item))
        self.waiting[index] = item
        if not self.include_fitted and isinstance(item, ProcessedScan):
            item.peak_set = []

    def consume(self, timeout=10):
        """Fetches the next work item from the input
        queue :attr:`queue`, blocking for at most `timeout` seconds.

        Parameters
        ----------
        timeout : int, optional
            The duration to allow the process to block
            for while awaiting new work items.

        Returns
        -------
        bool
            Whether or not a new work item was found waiting
            on the :attr:`queue`
        """
        blocking = timeout != 0
        try:
            item, index, _ = self.queue.get(blocking, timeout)
            self.queue.task_done()
            # DONE message may be sent many times.
            while item == DONE:
                item, index, _ = self.queue.get(blocking, timeout)
                self.queue.task_done()
            self.store_item(item, index)
            return True
        except QueueEmpty:
            return False

    def start_helper_producers(self):
        """Starts the additional :class:`ScanTransformingProcess` workers
        in :attr:`helper_producers` if they have not been started already.

        Should only be invoked once
        """
        if self.started_helpers:
            return
        self.started_helpers = True
        for helper in self.helper_producers:
            if helper.is_alive():
                continue
            helper.start()

    def produce(self, scan):
        """Performs any final quality controls on the outgoing
        :class:`ProcessedScan` object and takes care of any internal
        details.

        Resets :attr:`count_since_last` to `0`.

        Parameters
        ----------
        scan : ProcessedScan
            The scan object being finalized for hand-off
            to client code

        Returns
        -------
        ProcessedScan
            The version of `scan` ready to be used by other
            parts of the program
        """
        self.count_since_last = 0
        return scan

    def count_pending_items(self):
        return len(self.waiting)

    def drain_queue(self):
        i = 0
        while self.count_pending_items() < 500 and self.consume(0):
            self.count_jobs_done += 1
            i += 1
        return i

    def print_state(self):
        try:
            if self.queue.qsize() > 0:
                self.log("%d since last work item" % (self.count_since_last,))
                keys = sorted(self.waiting.keys())
                if len(keys) > 5:
                    self.log("Waiting Keys: %r..." % (keys[:5],))
                else:
                    self.log("Waiting Keys: %r" % (keys,))
                self.log("%d Keys Total" % (len(self.waiting),))
                self.log("The last index handled: %r" % (self.last_index,))
                self.log("Number of items waiting in the queue: %d" %
                         (self.queue.qsize(),))
        except NotImplementedError:
            # Some platforms do not support qsize
            pass
        for worker in ([self.primary_worker] + list(self.helper_producers)):
            code = worker.exitcode
            if code is not None and code != 0:
                self.log("%r has exit code %r" % (worker, code))
                worker.join(5)

    def __iter__(self):
        has_more = True
        # Log the state of the collator every 3 minutes
        status_monitor = CallInterval(60 * 3, self.print_state)
        status_monitor.start()
        while has_more:
            if self.consume(1):
                self.count_jobs_done += 1
                try:
                    if self.queue.qsize() > 500:
                        self.drain_queue()
                except NotImplementedError:
                    # Some platforms do not support qsize
                    pass
            if self.last_index is None:
                keys = sorted(self.waiting)
                if keys:
                    i = 0
                    n = len(keys)
                    found_content = False
                    while i < n:
                        scan = self.waiting.pop(keys[i])
                        if scan == SCAN_STATUS_SKIP:
                            self.last_index = keys[i]
                            i += 1
                            continue
                        else:
                            found_content = True
                            break
                    if found_content:
                        self.last_index = scan.index
                        yield self.produce(scan)
                    if self.last_index is not None:
                        self.start_helper_producers()
            elif self.last_index + 1 in self.waiting:
                while self.last_index + 1 in self.waiting:
                    scan = self.waiting.pop(self.last_index + 1)
                    if scan == SCAN_STATUS_SKIP:
                        self.last_index += 1
                        continue
                    else:
                        self.last_index = scan.index
                        yield self.produce(scan)
            elif len(self.waiting) == 0:
                if self.all_workers_done():
                    self.log("All Workers Claim Done.")
                    has_something = self.consume()
                    self.log("Checked Queue For Work: %r" % has_something)
                    if not has_something and len(self.waiting) == 0 and self.queue.empty():
                        has_more = False
            else:
                self.count_since_last += 1
                if self.count_since_last % 1000 == 0:
                    self.print_state()
        status_monitor.stop()



import unittest

from ms_deisotope.data_source import MGFLoader, Scan
from ms_deisotope.test.common import datafile
from ms_deisotope.data_source import infer_type


class TestMGFLoaderScanBehavior(unittest.TestCase):
    path = datafile("small.mgf")

    @property
    def reader(self):
        return infer_type.MSFileLoader(self.path)

    def test_index(self):
        reader = self.reader
        assert len(reader.index) == 34
        scan = reader.get_scan_by_id(
            'small.10.10')
        assert scan.id ==\
            'small.10.10'
        scan = reader[10]
        assert scan.index == 10

    def test_get_time(self):
        reader = self.reader
        scan = reader.get_scan_by_time(0.3)
        assert scan.id == 'small.31.31'
        scan = next(reader.start_from_scan(rt=0.3, grouped=False))
        assert scan.id == 'small.31.31'

    def test_annotations(self):
        scan = self.reader[10]
        assert scan.annotations == {}


    def test_scan_interface(self):
        reader = self.reader
        scan = next(reader)
        assert isinstance(scan, Scan)
        assert not scan.is_profile
        assert scan.precursor_information.precursor_scan_id is None


if __name__ == '__main__':
    unittest.main()



'''A collection of different strategies for iterating over streams
of :class:`~.Scan`-like objects.
'''
from . import ScanBunch


def _null_scan_validator(scan):
    return True


def _null_scan_cacher(scan):
    return None


ITERATION_MODE_GROUPED = "grouped"
ITERATION_MODE_SINGLE = "single"


class _ScanIteratorImplBase(object):
    """Internal base class for scan iteration strategies

    Attributes
    ----------
    iterator : :class:`Iterator`
        An iterator that produces a raw scan which will be packaged by :attr:`scan_packer`
    scan_cacher : :class:`Callable`
        A callable that will cache the scan, or a no-op
    scan_packer : :class:`Callable`
        A callable that will package a raw scan object into a :class:`Scan` object
    scan_validator : :class:`Callable`
        A callable that will check if a raw scan object is valid or not for filtering
        out unwanted entries, or a no-op
    """

    def __init__(self, iterator, scan_packer, scan_validator=None, scan_cacher=None):
        if scan_validator is None:
            scan_validator = _null_scan_validator
        if scan_cacher is None:
            scan_cacher = _null_scan_cacher
        self.iterator = iterator
        self.scan_packer = scan_packer
        self.scan_validator = scan_validator
        self.scan_cacher = scan_cacher
        self._producer = None

    def __iter__(self):
        return self

    def next(self):
        if self._producer is None:
            self._producer = self._make_producer()
        return next(self._producer)

    def __next__(self):
        return self.next()

    def _make_producer(self):
        raise NotImplementedError()

    @property
    def iteration_mode(self):
        """The iteration mode of the strategy, "grouped" or "single"
        """
        raise NotImplementedError()

    @classmethod
    def from_scan_source(cls, iterator, scan_source):
        """Create an iterator strategy for `iterator` from `scan_source`

        Parameters
        ----------
        iterator : :class:`Iterable`
            The iterator over raw scan data to consume
        scan_source : :class:`~.ScanIterator`
            The data extraction wrapper to provide with an iteration strategy
        """
        return cls(iterator, scan_source._make_scan, scan_source._validate, scan_source._cache_scan)



class _SingleScanIteratorImpl(_ScanIteratorImplBase):
    """Iterate over individual scans.

    The default strategy when MS1 scans are missing.
    """

    @property
    def iteration_mode(self):
        return ITERATION_MODE_SINGLE

    def _make_producer(self):
        _make_scan = self.scan_packer
        _validate = self.scan_validator
        _cache_scan = self.scan_cacher
        for scan in self.iterator:
            packed = _make_scan(scan)
            if not _validate(packed):
                continue
            _cache_scan(packed)
            yield packed


class _FakeGroupedScanIteratorImpl(_SingleScanIteratorImpl):
    '''Mimics the interface of :class:`_GroupedScanIteratorImpl` for
    scan sequences which only support single scans, or which do not
    guarantee sequential access to precursor/product collections.
    '''
    @property
    def iteration_mode(self):
        return ITERATION_MODE_GROUPED

    def _make_producer(self):
        generator = super(_FakeGroupedScanIteratorImpl, self)._make_producer()
        for scan in generator:
            if scan.ms_level == 1:
                yield ScanBunch(scan, [])
            else:
                yield ScanBunch(None, [scan])


class _GroupedScanIteratorImpl(_ScanIteratorImplBase):
    """Iterate over related scan bunches.

    The default strategy when MS1 scans are known to be
    present, even if MSn scans are not.
    """

    @property
    def iteration_mode(self):
        return ITERATION_MODE_GROUPED

    def _make_producer(self):
        _make_scan = self.scan_packer
        _validate = self.scan_validator
        _cache_scan = self.scan_cacher

        precursor_scan = None
        product_scans = []

        current_level = 1

        for scan in self.iterator:
            packed = _make_scan(scan)
            if not _validate(packed):
                continue
            _cache_scan(packed)
            if packed.ms_level > 1:
                # inceasing ms level
                if current_level < packed.ms_level:
                    current_level = packed.ms_level
                # decreasing ms level
                elif current_level > packed.ms_level:
                    current_level = packed.ms_level
                product_scans.append(packed)
            elif packed.ms_level == 1:
                if current_level > 1:
                    precursor_scan.product_scans = list(product_scans)
                    yield ScanBunch(precursor_scan, product_scans)
                else:
                    if precursor_scan is not None:
                        precursor_scan.product_scans = list(product_scans)
                        yield ScanBunch(precursor_scan, product_scans)
                precursor_scan = packed
                product_scans = []
            else:
                raise ValueError("Could not interpret MS Level %r" % (packed.ms_level,))
        if precursor_scan is not None:
            yield ScanBunch(precursor_scan, product_scans)



# cython: embedsignature=True
cimport cython

from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM

from ms_deisotope._c.feature_map.lcms_feature cimport LCMSFeature
from ms_deisotope._c.averagine cimport (
    neutral_mass as calc_neutral_mass, TheoreticalIsotopicPattern)
from ms_deisotope._c.peak_set cimport DeconvolutedPeak

cimport numpy as np
import numpy as np

np.import_array()



@cython.freelist(1000000)
cdef class map_coord(object):
    def __init__(self, mz, time):
        self.mz = mz
        self.time = time
    
    def __getitem__(self, i):
        if i == 0:
            return self.mz
        elif i == 1:
            return self.time
        else:
            raise IndexError(i)

    def __iter__(self):
        yield self.mz
        yield self.time

    def __reduce__(self):
        return map_coord, (self.mz, self.time,)

    def __repr__(self):
        return "map_coord(mz=%0.4f, time=%0.4f)" % (self.mz, self.time)


@cython.freelist(10000000)
cdef class LCMSFeatureSetFit(object):
    def __init__(self, features, theoretical, score, charge,
                 missing_features=0, supporters=None, data=None,
                 neutral_mass=None, n_points=0,
                 scores=None, times=None):
        if supporters is None:
            supporters = []
        if scores is None:
            scores = np.array([], dtype=np.float64)
        if times is None:
            times = np.array([], dtype=np.float64)
        self.features = features
        self.theoretical = theoretical
        self.score = score
        self.charge = charge
        self.data = data
        self.missing_features = missing_features
        self.monoisotopic_feature = features[0]
        self.supporters = supporters
        self.mz = self.theoretical.get_monoisotopic_mz()
        if neutral_mass is None:
            neutral_mass = calc_neutral_mass(self.mz, self.charge)
        self.neutral_mass = neutral_mass
        self.n_points = n_points
        self.scores = scores
        self.times = times

    @staticmethod
    cdef LCMSFeatureSetFit _create(list features, TheoreticalIsotopicPattern theoretical,
                                   double score, int charge, size_t missing_features,
                                   list supporters, object data, double neutral_mass,
                                   size_t n_points, np.ndarray scores, np.ndarray times):
        cdef:
            LCMSFeatureSetFit inst
        inst = LCMSFeatureSetFit.__new__(LCMSFeatureSetFit)
        if supporters is None:
            supporters = []
        if scores is None:
            scores = np.array([], dtype=np.float64)
        if times is None:
            times = np.array([], dtype=np.float64)

        inst.features = features
        inst.theoretical = theoretical
        inst.score = score
        inst.charge = charge
        inst.data = data
        inst.missing_features = missing_features
        inst.monoisotopic_feature = features[0]
        inst.supporters = supporters
        inst.mz = theoretical.get_monoisotopic_mz()
        if neutral_mass is None:
            neutral_mass = calc_neutral_mass(inst.mz, inst.charge)
        inst.neutral_mass = neutral_mass
        inst.n_points = n_points
        inst.scores = scores
        inst.times = times
        return inst

    def clone(self):
        return self.__class__(
            self.features, self.theoretical, self.score, self.charge,
            self.missing_features, self.supporters, self.data,
            self.neutral_mass, self.n_points, self.scores, self.times)

    def __reduce__(self):
        return self.__class__, (
            self.features, self.theoretical, self.score, self.charge,
            self.missing_features, self.supporters, self.data, self.neutral_mass,
            self.n_points, self.scores, self.times)

    cpdef bint _eq(self, LCMSFeatureSetFit other):
        cdef bint val
        val = (self.score == other.score and
               self.charge == other.charge and
               self.features == other.features and
               self.theoretical == other.theoretical)
        if self.data is not None or other.data is not None:
            val = val & (self.data == other.data)
        return val

    cpdef bint _ne(self, LCMSFeatureSetFit other):
        return not (self == other)

    cpdef bint _lt(self, LCMSFeatureSetFit other):
        return self.score < other.score

    cpdef bint _gt(self, LCMSFeatureSetFit other):
        return self.score > other.score

    def __richcmp__(self, LCMSFeatureSetFit other, int code):
        if other is None:
            if code == 3:
                return True
            else:
                return False

        if code == 0:
            return self._lt(other)
        elif code == 2:
            return self._eq(other)
        elif code == 3:
            return not (self._eq(other))
        elif code == 4:
            return self._gt(other)

    def __hash__(self):
        return hash((self.monoisotopic_feature.mz, self.charge))

    def __iter__(self):
        return iter(self.features)

    def __len__(self):
        return len(self.features)

    @property
    def npeaks(self):
        return len(self)

    def __repr__(self):
        return "LCMSFeatureSetFit(score=%0.5f, charge=%d, size=%d, monoisotopic_mz=%0.5f, %0.2f-%0.2f)" % (
            self.score, self.charge, len(self), self.monoisotopic_feature.mz,
            self.start.time, self.end.time)

    @property
    def start(self):
        first = self.features[0]
        if first is None:
            raise Exception()
        return map_coord(first.mz, first.start_time)

    @property
    def end(self):
        for last in reversed(self.features):
            if last is None:
                continue
            return map_coord(last.mz, last.end_time)



# cython: embedsignature=True
# 

from libc.math cimport sqrt, exp, pi

cimport cython
cimport numpy as np
import numpy as np


from ms_deisotope._c.feature_map.lcms_feature cimport LCMSFeature
from ms_deisotope._c.feature_map.shape_fitter cimport (
    gaussian_shape,
    bigaussian_shape,
    skewed_gaussian_shape)


np.import_array()



cpdef object find_immediate_minima(np.ndarray[double, ndim=1] x, double scale):
    cdef:
        size_t i, n
        list maxima, minima
        double val, prev, nxt

    maxima = []
    minima = []
    n = x.shape[0]
    for i in range(1, n - 1):
        val = x[i]
        prev = x[i - 1]
        nxt = x[i + 1]
        if val > prev * (1 + scale) and val > nxt * (1 + scale):
            maxima.append(i)
        elif val < prev * scale and val < nxt * scale:
            minima.append(i)

    return np.array(maxima, dtype=np.int64), np.array(minima, dtype=np.int64)


cpdef object locate_extrema(np.ndarray[double, ndim=1] x):
    cdef:
        size_t i, n
        list maxima, minima
        double val, prev, nxt

    maxima = []
    minima = []
    n = x.shape[0]
    for i in range(1, n - 1):
        val = x[i]
        prev = x[i - 1]
        nxt = x[i + 1]
        if val > prev and val > nxt:
            maxima.append(i)
        elif val < prev and val < nxt:
            minima.append(i)

    return np.array(maxima, dtype=np.int64), np.array(minima, dtype=np.int64)


def interpolate(xs, ys, n=200):
    new_xs = np.linspace(xs.min(), xs.max(), n)
    new_ys = np.interp(new_xs, xs, ys)
    return new_xs, new_ys


cdef object sliding_mean(np.ndarray[double, ndim=1, mode='c'] ys):
    return (np.concatenate((ys[1:], [0])) + ys + np.concatenate(([0], ys[:-1]))) / 3


cdef object sliding_median(np.ndarray[double, ndim=1, mode='c'] ys):
    cdef:
        np.ndarray[double, ndim=1, mode='c'] arr
        size_t i, n
        double value
    arr = np.zeros_like(ys)
    n = len(ys)
    for i in range(n):
        if i == 0:
            value = np.median(np.concatenate(([0], ys[:2]),))
        elif i == n - 1:
            value = np.median(np.concatenate(([0], ys[n - 2:]),))
        else:
            value = np.median(ys[i - 1:i + 2])
        arr[i] = value
    return arr


cpdef size_t binsearch(np.ndarray[double, ndim=1, mode='c'] array, double value):
    cdef:
        size_t lo, hi, mid
        double point
    lo = 0
    hi = array.shape[0]
    while hi != lo:
        mid = (hi + lo) / 2
        point = array[mid]
        if value == point:
            return mid
        elif hi - lo == 1:
            return mid
        elif point > value:
            hi = mid
        else:
            lo = mid


cdef np.ndarray[double, ndim=1, mode='c'] gauss(np.ndarray[double, ndim=1, mode='c'] x,
                                                double mean, double variance, double scale=2.0):
    return 1. / np.sqrt(2 * np.pi * variance) * np.exp(-(x - mean) ** 2 / (scale * variance))


cpdef np.ndarray[double, ndim=1, mode='c'] gaussian_smooth(
        np.ndarray[double, ndim=1, mode='c'] x,
        np.ndarray[double, ndim=1, mode='c'] y, double width=0.05):
    cdef:
        np.ndarray[double, ndim=1, mode='c'] smoothed, x_slice, y_slice, weights
        size_t i, n, low_edge, high_edge
        double center, spread

    smoothed = np.zeros_like(y)

    n = x.shape[0]
    for i in range(n):
        low_edge = binsearch(x, x[i] - width)
        high_edge = binsearch(x, x[i] + width)
        if high_edge - 1 == low_edge or high_edge == low_edge:
            smoothed[i] = y[i]
            continue
        x_slice = x[low_edge:high_edge]
        y_slice = y[low_edge:high_edge]
        center = np.mean(x_slice)
        spread = np.var(x_slice)
        weights = gauss(x_slice, center, spread)
        smoothed[i] = ((y_slice * weights).sum() / weights.sum())
    return smoothed


cpdef np.ndarray[double, ndim=1, mode='c'] smooth_leveled(
        np.ndarray[double, ndim=1, mode='c'] xs,
        np.ndarray[double, ndim=1, mode='c'] ys, double level=0):
    if level == 0:
        return ys
    elif level == 1:
        return sliding_mean(ys)
    elif level == 2:
        return sliding_mean(sliding_median(ys))
    elif level == 3:
        return gaussian_smooth(xs, ys, 0.05)
    else:
        return gaussian_smooth(xs, ys, level)


cdef class PeakBoundary(object):
    cdef:
        public double first_minimum
        public double maximum
        public double second_minimum
        public double first_minimum_index
        public double maximum_index
        public double second_minimum_index
        public double total_distance

    def __init__(self, first_minimum, maximum, second_minimum,
                 first_minimum_index, maximum_index, second_minimum_index):
        self.first_minimum = first_minimum
        self.maximum = maximum
        self.second_minimum = second_minimum
        self.first_minimum_index = first_minimum_index
        self.maximum_index = maximum_index
        self.second_minimum_index = second_minimum_index
        self.total_distance = self.compute_distance()

    cdef double compute_distance(self):
        return (self.maximum - self.first_minimum) + (self.maximum - self.second_minimum)

    def __repr__(self):
        return "PeakBoundary(%0.2f, %0.2f, %0.2f, %0.3e)" % (
            self.first_minimum_index, self.maximum_index, self.second_minimum_index,
            self.total_distance)

    def split(self, profile):
        first_minimum = profile.nodes.find_time(self.first_minimum_index)[1]
        second_minimum = profile.nodes.find_time(self.second_minimum_index)[1]
        if first_minimum == 0:
            first_minimum += 1
        prefix = profile[:first_minimum]
        peak = profile[(first_minimum - 1):second_minimum + 1]
        suffix = profile[second_minimum + 1:]
        cases = LCMSFeature(prefix), LCMSFeature(peak), LCMSFeature(suffix)
        return cases


cdef class ValleyPoint(object):
    cdef:
        public double first_maximum
        public double minimum
        public double second_maximum
        public double first_maximum_index
        public double minimum_index
        public double second_maximum_index
        public double total_distance

    def __init__(self, first_maximum, minimum, second_maximum, first_maximum_index,
                 minimum_index, second_maximum_index):
        self.first_maximum = first_maximum
        self.minimum = minimum
        self.second_maximum = second_maximum
        self.first_maximum_index = first_maximum_index
        self.minimum_index = minimum_index
        self.second_maximum_index = second_maximum_index
        self.total_distance = self.compute_distance()

    cdef double compute_distance(self):
        return (self.first_maximum - self.minimum) + (self.second_maximum - self.minimum)

    def __repr__(self):
        return "ValleyPoint(%0.2f, %0.2f, %0.2f, %0.3e)" % (
            self.first_maximum_index, self.minimum_index, self.second_maximum_index, self.total_distance)

    def split(self, profile):
        index = profile.nodes.find_time(self.minimum_index)[1]
        before = LCMSFeature(profile[:index])
        after = LCMSFeature(profile[index:])
        return before, after


cdef class ProfileSplitter(object):
    cdef:
        public object profile
        public list partition_sites
        public np.ndarray xs
        public np.ndarray ys

    def __init__(self, profile):
        if isinstance(profile, LCMSFeature):
            self.profile = profile
            self.xs, self.ys = profile.as_arrays()
        else:
            self.xs, self.ys = profile
            self.profile = None
        self.partition_sites = []

    def _extreme_indices(self, ys):
        maxima_indices, minima_indices = locate_extrema(ys)
        return maxima_indices.astype(np.int64), minima_indices.astype(np.int64)

    cpdef object interpolate(self, xs, ys, int n=200):
        new_xs = np.linspace(xs.min(), xs.max(), n)
        new_ys = np.interp(new_xs, xs, ys)
        return new_xs, new_ys

    @cython.boundscheck(False)
    def locate_peak_boundaries(self, double smooth=1, int interpolate_past=200):
        cdef:
            np.ndarray[double, ndim=1] xs, ys
            np.ndarray[np.int64_t, ndim=1] maxima_indices, minima_indices
            list candidates
            size_t i, j, k, max_i, max_j, min_k
            double y_i, y_j, y_k
            PeakBoundary point

        xs = self.xs
        ys = self.ys
        ys = sliding_mean(ys)
        if len(xs) > interpolate_past:
            xs, ys = self.interpolate(xs, ys, interpolate_past)
        if smooth:
            ys = smooth_leveled(xs, ys, smooth)

        maxima_indices, minima_indices = self._extreme_indices(ys)
        candidates = []
        for i in range(len(minima_indices)):
            min_i = minima_indices[i]
            for j in range(i + 1, len(minima_indices)):
                min_j = minima_indices[j]
                for k in range(len(maxima_indices)):
                    max_k = maxima_indices[k]
                    y_i = ys[min_i]
                    y_j = ys[min_j]
                    y_k = ys[max_k]
                    if min_i < max_k < min_j and (y_k - y_i) > (y_k * 0.01) and (
                            y_k - y_j) > (y_k * 0.01):
                        point = PeakBoundary(y_i, y_k, y_j, xs[min_i], xs[max_k], xs[min_j])
                        candidates.append(point)
        if candidates:
            candidates = sorted(candidates, key=lambda x: x.total_distance, reverse=True)
            best_point = candidates[0]
            self.partition_sites.append(best_point)

        return candidates

    @cython.boundscheck(False)
    def locate_valleys(self, double scale=0.3, double smooth=1, int interpolate_past=200):
        cdef:
            np.ndarray[double, ndim=1, mode='c'] xs
            np.ndarray[double, ndim=1, mode='c'] ys
            np.ndarray[np.int64_t, ndim=1, mode='c'] maxima_indices
            np.ndarray[np.int64_t, ndim=1, mode='c'] minima_indices
            list candidates
            size_t i, j, k, max_i, max_j, min_k, n_maxima, n_minima
            double y_i, y_j, y_k
            ValleyPoint point

        xs = self.xs
        ys = self.ys
        
        if len(xs) > interpolate_past:
            xs, ys = self.interpolate(xs, ys, interpolate_past)

        if smooth:
            ys = smooth_leveled(xs, ys, smooth)

        maxima_indices, minima_indices = self._extreme_indices(ys)
        candidates = []
        n_minima = len(minima_indices)
        n_maxima = len(maxima_indices)

        for i in range(n_maxima):
            max_i = maxima_indices[i]
            for j in range(i + 1, n_maxima):
                max_j = maxima_indices[j]
                for k in range(n_minima):
                    min_k = minima_indices[k]
                    y_i = ys[max_i]
                    y_j = ys[max_j]
                    y_k = ys[min_k]
                    if max_i < min_k < max_j and (y_i - y_k) > (y_i * scale) and (
                            y_j - y_k) > (y_j * scale):
                        point = ValleyPoint(y_i, y_k, y_j, xs[max_i], xs[min_k], xs[max_j])
                        candidates.append(point)
        if candidates:
            candidates = sorted(candidates, key=lambda x: x.total_distance, reverse=True)
            best_point = candidates[0]
            self.partition_sites.append(best_point)

        return candidates

    def extract_peak(self, peak_boundary):
        if self.profile is not None:
            cases = peak_boundary.split(self.profile)
        else:
            cases = None
        return cases

    def split_valley(self, valley):
        if self.profile is not None:
            return valley.split(self.profile)
        else:
            return None


def split_valleys(profiles, scale=0.3, n_levels=2):
    for i in range(n_levels):
        out = []
        for case in profiles:
            ps = ProfileSplitter(case)
            vallies = ps.locate_valleys(scale)
            if vallies:
                splitter = vallies[0]
                out.extend(p for p in ps.split_valley(splitter) if len(p) > 0)
            else:
                out.append(case)
        profiles = out
    return profiles



"""Defines types for describing mass spectrometry data transformations,
data processing sequences, and a database of controlled vocabulary terms
for them.

.. note::
    The entities in this module are only for *naming* these transformations
    and how they were combined to process mass spectra. For implementations,
    see other features of :mod:`ms_deisotope` and :mod:`ms_peak_picker`, or
    other libraries.
"""

from collections import OrderedDict

from six import string_types as basestring

from ms_deisotope.utils import uid

from .cv import Term, TermSet
# from .software import Software, SoftwareName


class DataTransformation(Term):
    """Describes a named data transformation, either
    using a controlled-vocabulary term or user-defined name.

    A :class:`DataTransformation` is equal to its name and its controlled
    vocabulary identifier.
    """
    pass


data_transformations = []


# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('data transformation', term_cls_name="DataTransformation", writer=cog.out)
# ]]]
data_transformations = TermSet([
    DataTransformation(u'file format conversion', u'MS:1000530',
                       (u'Conversion of one file format to another.'),
                       'data transformation',
                       [u'data transformation']),
    DataTransformation(u'data processing action', u'MS:1000543',
                       (u'Data processing attribute used to describe the type of data'
                        u'processing performed on the data file.'),
                       'data transformation',
                       [u'data transformation']),
    DataTransformation(u'Conversion to mzML', u'MS:1000544',
                       (u'Conversion of a file format to Proteomics Standards'
                        u'Initiative mzML file format.'),
                       'data transformation',
                       [u'file format conversion', u'data transformation']),
    DataTransformation(u'Conversion to mzMLb', u'MS:1002839',
                       (u'Conversion of a file format to Proteomics Standards'
                        u'Initiative mzMLb file format.'),
                       'data transformation',
                       [u'file format conversion', u'data transformation']),
    DataTransformation(u'Conversion to dta', u'MS:1000741',
                       (u'Conversion to dta format.'),
                       'data transformation',
                       [u'file format conversion', u'data transformation']),
    DataTransformation(u'Conversion to mzXML', u'MS:1000545',
                       (u'Conversion of a file format to Institute of Systems Biology'
                        u'mzXML file format.'),
                       'data transformation',
                       [u'file format conversion', u'data transformation']),
    DataTransformation(u'Conversion to mzData', u'MS:1000546',
                       (u'Conversion of a file format to Proteomics Standards'
                        u'Initiative mzData file format.'),
                       'data transformation',
                       [u'file format conversion', u'data transformation']),
    DataTransformation(u'charge deconvolution', u'MS:1000034',
                       (u'The determination of the mass of an ion based on the mass'
                        u'spectral peaks that represent multiple-charge ions.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'peak picking', u'MS:1000035',
                       (u'Spectral peak processing conducted on the acquired data to'
                        u'convert profile data to centroided data.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'deisotoping', u'MS:1000033',
                       (u'The removal of isotope peaks to represent the fragment ion'
                        u'as one data point and is commonly done to reduce complexity.'
                        u'It is done in conjunction with the charge state'
                        u'deconvolution.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'baseline reduction', u'MS:1000593',
                       (u'A process of removal of varying intensities generated due to'
                        u'variable energy absorption before further processing can'
                        u'take place. Baseline reduction facilitates meaningful'
                        u'comparision between intensities of m/z values.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'smoothing', u'MS:1000592',
                       (u'A process of reducing spikes of intensity in order to reduce'
                        u'noise while preserving real peak signal. Many algorithms can'
                        u'be applied for this process.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'precursor recalculation', u'MS:1000780',
                       (u'A process that recalculates existing precursor selected ions'
                        u'with one or more algorithmically determined precursor'
                        u'selected ions.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'data filtering', u'MS:1001486',
                       (u'Filtering out part of the data.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'retention time alignment', u'MS:1000745',
                       (u'The correction of the spectrum scan times, as used e.g. in'
                        u'label-free proteomics.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'intensity normalization', u'MS:1001484',
                       (u'Normalization of data point intensities.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'm/z calibration', u'MS:1001485',
                       (u'Calibration of data point m/z positions.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'charge state calculation', u'MS:1000778',
                       (u"A process that infers the charge state of an MSn spectrum's"
                        u'precursor(s) by the application of some algorithm.'),
                       'data transformation',
                       [u'data processing action', u'data transformation']),
    DataTransformation(u'area peak picking', u'MS:1000801',
                       (u'Spectral peak processing conducted on the acquired data to'
                        u'convert profile data to centroided data. The area defined by'
                        u'all raw data points that belong to the peak is reported.'),
                       'data transformation',
                       [u'peak picking', u'data processing action', u'data transformation']),
    DataTransformation(u'height peak picking', u'MS:1000802',
                       (u'Spectral peak processing conducted on the acquired data to'
                        u'convert profile data to centroided data. The maximum'
                        u'intensity of all raw data points that belong to the peak is'
                        u'reported.'),
                       'data transformation',
                       [u'peak picking', u'data processing action', u'data transformation']),
    DataTransformation(u'top hat baseline reduction', u'MS:1001994',
                       (u'Top-hat morphological filter based on the basic'
                        u"morphological operations 'erosion' and 'dilatation'."),
                       'data transformation',
                       [u'baseline reduction', u'data processing action', u'data transformation']),
    DataTransformation(u'convex hull baseline reduction', u'MS:1001995',
                       (u'Constructs the baseline by fitting multiple parabolas to the'
                        u'spectrum starting with the large scale structures.'),
                       'data transformation',
                       [u'baseline reduction', u'data processing action', u'data transformation']),
    DataTransformation(u'median baseline reduction', u'MS:1001996',
                       (u'The spectrum that will be baseline subtracted is divided'
                        u'into a number of segments.'),
                       'data transformation',
                       [u'baseline reduction', u'data processing action', u'data transformation']),
    DataTransformation(u'wavelet transformation smoothing', u'MS:1001997',
                       (u'The random noise is removed by using the undecimated wavelet'
                        u'transform." [DOI:10.1093/bioinformatics/btl355'),
                       'data transformation',
                       [u'smoothing', u'data processing action', u'data transformation']),
    DataTransformation(u'Gaussian smoothing', u'MS:1000784',
                       (u'Reduces intensity spikes by convolving the data with a one-'
                        u'dimensional Gaussian function.'),
                       'data transformation',
                       [u'smoothing', u'data processing action', u'data transformation']),
    DataTransformation(u'moving average smoothing', u'MS:1000785',
                       (u'Reduces intensity spikes by averaging each point with two or'
                        u'more adjacent points. The more adjacent points that used,'
                        u'the stronger the smoothing effect.'),
                       'data transformation',
                       [u'smoothing', u'data processing action', u'data transformation']),
    DataTransformation(u'Savitzky-Golay smoothing', u'MS:1000782',
                       (u'Reduces intensity spikes by applying local polynomial'
                        u'regression (of degree k) on a distribution (of at least k+1'
                        u'equally spaced points) to determine the smoothed value for'
                        u'each point. It tends to preserve features of the'
                        u'distribution such as relative maxima, minima and width,'
                        u"which are usually 'flattened' by other adjacent averaging"
                        u'techniques.'),
                       'data transformation',
                       [u'smoothing', u'data processing action', u'data transformation']),
    DataTransformation(u'LOWESS smoothing', u'MS:1000783',
                       (u'Reduces intensity spikes by applying a modelling method'
                        u'known as locally weighted polynomial regression. At each'
                        u'point in the data set a low-degree polynomial is fit to a'
                        u'subset of the data, with explanatory variable values near'
                        u'the point whose response is being estimated. The polynomial'
                        u'is fit using weighted least squares, giving more weight to'
                        u'points near the point whose response is being estimated and'
                        u'less weight to points further away. The value of the'
                        u'regression function for the point is then obtained by'
                        u'evaluating the local polynomial using the explanatory'
                        u'variable values for that data point. The LOESS fit is'
                        u'complete after regression function values have been computed'
                        u'for each of the n data points. Many of the details of this'
                        u'method, such as the degree of the polynomial model and the'
                        u'weights, are flexible.'),
                       'data transformation',
                       [u'smoothing', u'data processing action', u'data transformation']),
    DataTransformation(u'msPrefix precursor recalculation', u'MS:1000781',
                       (u'Recalculates one or more precursor selected ions by peak'
                        u'detection in the isolation windows of high accuracy MS'
                        u'precursor scans.'),
                       'data transformation',
                       [u'precursor recalculation', u'data processing action', u'data transformation']),
    DataTransformation(u'low intensity data point removal', u'MS:1000594',
                       (u'The removal of very low intensity data points that are'
                        u'likely to be spurious noise rather than real signal.'),
                       'data transformation',
                       [u'data filtering', u'data processing action', u'data transformation']),
    DataTransformation(u'high intensity data point removal', u'MS:1000746',
                       (u'The removal of very high intensity data points.'),
                       'data transformation',
                       [u'data filtering', u'data processing action', u'data transformation']),
    DataTransformation(u'area normalization', u'MS:1001999',
                       (u'Normalization of areas below the curves.'),
                       'data transformation',
                       [u'intensity normalization', u'data processing action', u'data transformation']),
    DataTransformation(u'below precursor intensity dominance charge state calculation', u'MS:1000779',
                       (u'Infers charge state as single or ambiguously multiple by'
                        u'determining the fraction of intensity below the precursor'
                        u'm/z.'),
                       'data transformation',
                       [u'charge state calculation', u'data processing action', u'data transformation']),
    DataTransformation(u'sophisticated numerical annotation procedure', u'MS:1001998',
                       (u'It searches for known patterns in the measured spectrum."'
                        u'[DOI:10.1021/ac951158i'),
                       'data transformation',
                       [u'area peak picking', u'peak picking', u'data processing action', u'data transformation']),
])
# [[[end]]]


def data_transformation(name):
    '''Translate a given name or identifier into a :class:`DataTransformation`
    instance.

    If no match is found in the database of known :class:`DataTransformation`
    types, a new dummy :class:`DataTransformation` is returned with all fields
    set to the value of ``name``

    Returns
    -------
    DataTransformation
    '''
    try:
        return data_transformations[name]
    except KeyError:
        return DataTransformation(name, name, name, name, [name])


class ProcessingMethod(object):
    '''Describes a single action applied to the associated spectra, composed of one
    or more data transformation operations.

    Attributes
    ----------
    operations: :class:`~.OrderedDict`
        The data transformations applied, mapping transformation term to any
        parameter value or the empty string.
    software_id: str
        A reference to a particular version of software used to carry out the
        transformations.
    order: int
        The order in which this action was performed, in the context of the
        parent :class:`DataProcessingInformation` object

    See Also
    --------
    :class:`DataTransformation`
    '''

    def __init__(self, operations=None, software_id=None, order=0):
        self.operations = operations or OrderedDict()
        self.software_id = software_id
        self.order = order

    def __getitem__(self, i):
        return list(self.operations.items())[i]

    def __len__(self):
        return len(self.operations)

    def __iter__(self):
        return iter(self.operations.items())

    def _get_names(self):
        return ["%r: %r" % (t.name, v) for t, v in self.operations.items()]

    def __repr__(self):
        template = "{self.__class__.__name__}({names}, order={self.order})"
        return template.format(self=self, names='[%s]' % ', '.join(self._get_names()))

    def add(self, operation, value=''):
        '''Add a new :class:`DataTransformation` operation to this method.

        Parameters
        ----------
        operation: :class:`DataTransformation` or :class:`str`
            The transformation applied. If passed as a :class:`str`, it will be translated
            into a :class:`DataTransformation`.
        value: :class:`object`
            A parameter value associated with the operation. Defaults to the empty string.
        '''
        if isinstance(operation, basestring):
            operation = data_transformation(operation)
        self.operations[operation] = value

    def update(self, operations):
        '''Add each operation in ``operations`` to this method.

        Parameters
        ----------
        operations: :class:`~.Mapping` or :class:`Iterable`
            If given a :class:`~.Mapping`, add each key-value
            pair using :meth:`add`. Otherwise extract keys or key-value pairs
            if :class:`Iterable` is supported, adding each in succession
        '''
        if isinstance(operations, dict):
            for op, val in operations.items():
                self.add(op, val)
        else:
            for op in operations:
                if isinstance(op, dict):
                    self.add(op['name'], op['value'])
                elif isinstance(op, (tuple, list)) and len(op) == 2:
                    self.add(*op)
                else:
                    self.add(op)


class DataProcessingInformation(object):
    '''A :class:`Sequence` of :class:`ProcessingMethod` instances, describing the sequence
    of data transformations applied to the spectra in the data associated with this instance.

    Attributes
    ----------
    methods: list
        A list of :class:`ProcessingMethod` instances describing the individual operations done
        in an ordered fashion.
    id: int
        A within-source unique identifier
    '''

    def __init__(self, methods=None, id=None):
        self.methods = methods or []
        self.id = id or uid()

    def __getitem__(self, i):
        return self.methods[i]

    def __len__(self):
        return len(self.methods)

    def __iter__(self):
        return iter(self.methods)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.methods}, id={self.id!r})"
        return template.format(self=self)


__all__ = [
    "DataTransformation", "data_transformations", 'data_transformation',
    "ProcessingMethod", "DataProcessingInformation"
]



import operator
import warnings

from collections import defaultdict

cimport cython

from cpython cimport Py_INCREF
from cpython.list cimport PyList_GET_ITEM, PyList_GET_SIZE, PyList_New, PyList_SET_ITEM
from cpython.tuple cimport PyTuple_GET_ITEM
from cpython.int cimport PyInt_AsLong
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem
from cpython.object cimport PyObject
from cpython.set cimport PySet_Add

from ms_peak_picker._c.peak_set cimport FittedPeak, PeakSet

from ms_deisotope._c.scoring cimport IsotopicFitRecord
from ms_deisotope._c.peak_dependency_network.intervals cimport SpanningMixin, IntervalTreeNode
from ms_deisotope._c.peak_dependency_network.subgraph cimport ConnectedSubgraph


cdef object score_getter = operator.attrgetter("score")


@cython.freelist(1000000)
cdef class PeakNode(object):
    """
    Represent a single FittedPeak and the junction of multiple
    IsotopicFitRecords which depend uponit

    Attributes
    ----------
    links : dict
        Mapping IsotopicFitRecords which depend upon this peak to those fit's scores
    peak : ms_peak_picker.FittedPeak
        The peak being depended upon
    """

    @staticmethod
    cdef PeakNode _create(FittedPeak peak):
        cdef PeakNode inst = PeakNode.__new__(PeakNode)
        inst.peak = peak
        inst.links = dict()
        inst._hash = hash(peak)
        return inst

    def __init__(self, peak, links=None):
        if links is None:
            links = {}
        self.peak = peak
        self.links = links
        self._hash = hash(self.peak)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        try:
            return self.peak == other.peak
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other

    def __contains__(self, fit):
        return fit in self.links

    def __repr__(self):
        return "PeakNode(%s, %s)" % (self.peak, self.links)


cdef class DependenceCluster(SpanningMixin):
    """
    Represent a cluster of isotopic fits which are overlapping

    Attributes
    ----------
    dependencies : set
        The set of all fits which are codependent
    start: float
        The smallest mz of all elements of this cluster
    end: float
        The largest mz of all elements of this cluster
    maximize: bool
        Whether the objective is to maximize or minimize the
        isotopic fit score
    """

    def __init__(self, parent=None, dependencies=None, maximize=True):
        if parent is None:
            parent = self
        if dependencies is None:
            dependencies = []
        else:
            dependencies = sorted(dependencies, key=lambda x: x.score, reverse=maximize)
        self.parent = parent
        self.dependencies = dependencies
        self.maximize = maximize
        self._reset()

    cpdef _reset(self):
        self.start = self._start()
        self.end = self._end()
        self.best_fit = self._best_fit()

    cpdef add(self, IsotopicFitRecord fit):
        """
        Adds a new IsotopicFitRecord to this cluster, and ensures the sorted
        property still holds

        Parameters
        ----------
        fit : IsotopicFitRecord
            The fit being added to the record

        """
        self.dependencies.append(fit)
        self.dependencies.sort(reverse=self.maximize)
        self._reset()

    cpdef disjoint_subset(self):
        graph = self.build_graph()
        return graph.find_heaviest_path()

    cpdef ConnectedSubgraph build_graph(self):
        graph = ConnectedSubgraph(self.dependencies, maximize=self.maximize)
        return graph

    cdef IsotopicFitRecord _best_fit(self):
        """
        Retrieve the absolute best single fit for the cluster

        Returns
        -------
        name : IsotopicFitRecord
            The best scoring IsotopicFitRecord in this cluster
        """
        return self.dependencies[0]

    cpdef list disjoint_best_fits(self):
        """
        Compute the best set of disjoint isotopic fits spanning this cluster

        Returns
        -------
        list of IsotopicFitRecord
        """
        fit_sets = tuple(self.disjoint_subset())
        best_fits = fit_sets
        return [node.fit for node in best_fits]

    cdef double _start(self):
        """
        Determine the first mz coordinate for members of this cluster

        Returns
        -------
        float
        """
        return min([member.experimental[0].mz for member in self.dependencies])

    cdef double _end(self):
        """
        Determines the last mz coordinate for members of this cluster

        Returns
        -------
        float
        """
        return max([member.experimental[-1].mz for member in self.dependencies])

    def __len__(self):
        return len(self.dependencies)

    def __iter__(self):
        return iter(self.dependencies)

    def __contains__(self, fit):
        return fit in self.dependencies

    def __eq__(self, other):
        return self.dependencies == other.dependencies

    def __repr__(self):
        return "DependenceCluster(dependencies=[%s], start=%0.4f, end=%0.4f)" % (
            '\n'.join(map(str, self.dependencies[:10])), self.start, self.end)

    def __getitem__(self, i):
        return self.dependencies[i]


cdef class PeakDependenceGraphBase(object):

    cpdef reset(self):
        # Keep a record of all clusters from previous iterations
        self._all_clusters.extend(
            self.clusters if self.clusters is not None else [])
        self.nodes = dict()
        self.dependencies = set()
        self._interval_tree = None
        self._populate_initial_graph()

    cpdef _populate_initial_graph(self):
        cdef:
            size_t i, n
            FittedPeak peak
        n = self.peaklist.get_size()
        for i in range(n):
            peak = self.peaklist.getitem(i)

        for peak in self.peaklist:
            self.nodes[peak.index] = PeakNode(peak)

    cpdef add_fit_dependence(self, IsotopicFitRecord fit_record):
        '''Add the relatoinship between the experimental peaks
        in `fit_record` to the graph, expressed as a hyper-edge
        denoted by `fit_record`.

        This adds `fit_record` to :attr:`PeakNode.links` for each
        node corresponding to the :class:`~.FittedPeak` instance in
        :attr:`IsotopicFitRecord.experimental` of `fit_record`. It also
        adds `fit_record to :attr:`dependencies`

        Parameters
        ----------
        fit_record: :class:`~.IsotopicFitRecord`
        '''
        cdef:
            size_t i, n
            FittedPeak peak
            PeakNode node
        n = PyList_GET_SIZE(fit_record.experimental)
        for i in range(n):
            peak = <FittedPeak>PyList_GET_ITEM(fit_record.experimental, i)
            if peak.index == 0:
                continue
            node = <PeakNode>PyDict_GetItem(self.nodes, peak.index)
            PyDict_SetItem(node.links, fit_record, fit_record.score)
        self.dependencies.add(fit_record)

    cpdef list nodes_for(self, IsotopicFitRecord fit_record, dict cache=None):
        cdef:
            list result
            size_t i, n
            FittedPeak p
            PeakNode node
        if cache is None:
            result = []
            n = PyList_GET_SIZE(fit_record.experimental)
            for i in range(n):
                p = <FittedPeak>PyList_GET_ITEM(fit_record.experimental, i)
                if p.peak_count >= 0:
                    node = <PeakNode>PyDict_GetItem(self.nodes, p.index)
                    result.append(node)
            return result
        else:
            try:
                return cache[fit_record]
            except KeyError:
                result = []
                n = PyList_GET_SIZE(fit_record.experimental)
                for i in range(n):
                    p = <FittedPeak>PyList_GET_ITEM(fit_record.experimental, i)
                    if p.peak_count >= 0:
                        node = <PeakNode>PyDict_GetItem(self.nodes, p.index)
                        result.append(node)
                PyDict_SetItem(cache, fit_record, result)
                return result



from brainpy._c.isotopic_distribution cimport TheoreticalPeak
from ms_peak_picker._c.peak_set cimport PeakSet, FittedPeak

from ms_deisotope._c.deconvoluter_base cimport DeconvoluterBase
from ms_deisotope._c.averagine cimport TheoreticalIsotopicPattern

ctypedef fused fit_collection:
    list
    set
    object


cdef double INFINITY = float('inf')


cdef class IsotopicFitRecord(object):
    cdef:
        public FittedPeak seed_peak
        public double score
        public int charge
        public list experimental
        public TheoreticalIsotopicPattern theoretical
        public FittedPeak monoisotopic_peak
        public int missed_peaks
        public object data
        public Py_hash_t _hash

    @staticmethod
    cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,
                                   list experimental, object data=*, int missed_peaks=*)

    cpdef bint _eq(self, IsotopicFitRecord other)
    cpdef bint _ne(self, IsotopicFitRecord other)
    cpdef bint _lt(self, IsotopicFitRecord other)
    cpdef bint _gt(self, IsotopicFitRecord other)


cdef class FitSelectorBase(object):
    cdef:
        public double minimum_score

    cpdef IsotopicFitRecord best(self, object results)
    cdef IsotopicFitRecord _best_from_set(self, set results)

    cpdef bint reject(self, IsotopicFitRecord result)
    cpdef bint reject_score(self, double score)
    cpdef bint is_maximizing(self)


cdef class IsotopicFitterBase(object):
    cdef:
        public FitSelectorBase select

    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected)
    cpdef bint reject(self, IsotopicFitRecord fit)
    cpdef bint reject_score(self, double score)
    cpdef bint is_maximizing(self)

    cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs)


cdef class LeastSquaresFitter(IsotopicFitterBase):
    pass


cdef class GTestFitter(IsotopicFitterBase):
    pass

cdef class ScaledGTestFitter(IsotopicFitterBase):
    pass

cdef class MSDeconVFitter(IsotopicFitterBase):
    cdef public double mass_error_tolerance

    cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=*, double minimum_signal_to_noise=*) nogil


cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):
    cdef:
        public double mass_error_tolerance
        public double penalty_factor


cdef class FunctionScorer(IsotopicFitterBase):
    cdef:
        object function


cdef class InterferenceDetection(object):
    cdef public PeakSet peaklist

    cdef double detect_interference(self, list experimental_peaks, double lower=*, double upper=*)


cdef class DistinctPatternFitter(IsotopicFitterBase):
    cdef:
        public InterferenceDetection interference_detector
        public ScaledGTestFitter g_test_scaled
        public double peak_count_scale
        public double domain_scale

cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):
    cdef:
        public double scale_factor
        public PenalizedMSDeconVFitter scorer

    cpdef double _calculate_scale_factor(self, PeakSet peaklist)
    cdef void scale_fitted_peaks(self, list experimental, double factor)
    cdef void scale_theoretical_peaks(self, list theoretical, double factor)


cdef class DotProductFitter(IsotopicFitterBase):
    pass



import unittest

from ms_deisotope.data_source import memory
from ms_deisotope.test.common import datafile
from ms_deisotope.data_source import infer_type

scan_ids = [
    "controllerType=0 controllerNumber=1 scan=10014",
    "controllerType=0 controllerNumber=1 scan=10015",
    "controllerType=0 controllerNumber=1 scan=10016"
]


class TestMemoryScanSource(unittest.TestCase):
    path = datafile("three_test_scans.mzML")

    @property
    def source_reader(self):
        return infer_type.MSFileLoader(self.path)

    @property
    def prepare_source(self):
        source = self.source_reader
        loader = memory.MemoryScanLoader.build(source)
        return loader

    def test_iteration(self):
        g = iter(scan_ids)
        bunch = next(self.prepare_source)
        assert bunch.precursor.id == next(g)
        for product in bunch.products:
            assert product.id == next(g)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
'''Implementations of averagine-based deconvoluters.

Averagine-based deconvoluters use an "average monomer" isotopic model
to interpolate the isotopic patterns for any peak in the experimental
spectrum. The term "averagine" comes from the name Senko gave to the
"average amino acid" when introducing the concept in [1].

References
----------
[1] Senko, M. W., Beu, S. C., & McLafferty, F. W. (1995). Determination of monoisotopic
    masses and ion populations for large biomolecules from resolved isotopic distributions.
    Journal of the American Society for Mass Spectrometry, 6(4), 229–233.
    http://doi.org/10.1016/1044-0305(95)00017-8
'''

from ms_deisotope.averagine import PROTON, AveragineCache, peptide, glycopeptide, glycan
from ms_deisotope.constants import IGNORE_BELOW, TRUNCATE_AFTER, SCALE_METHOD

from ms_deisotope.scoring import penalized_msdeconv

from .base import (
    DeconvoluterBase)

from .exhaustive import (
    ExhaustivePeakSearchDeconvoluterBase,
    PeakDependenceGraphDeconvoluterBase)

from .utils import count_placeholders, prepare_peaklist


class AveragineDeconvoluterBase(DeconvoluterBase):
    """A base class derived from :class:`DeconvoluterBase` which provides some common methods
    for fitting isotopic patterns using an Averagine model.
    """

    def __init__(self, use_subtraction=False, scale_method=SCALE_METHOD, merge_isobaric_peaks=True,
                 minimum_intensity=5., *args, **kwargs):
        super(AveragineDeconvoluterBase, self).__init__(
            use_subtraction, scale_method, merge_isobaric_peaks,
            minimum_intensity, *args, **kwargs)

    def fit_theoretical_distribution(self, peak, error_tolerance, charge, charge_carrier=PROTON, truncate_after=0.8,
                                     ignore_below=IGNORE_BELOW):
        """Fit an isotopic pattern seeded at `peak` at `charge` charge.

        Generates a theoretical isotopic pattern using :attr:`averagine`, calls
        :meth:`match_theoretical_isotopic_distribution`
        to extract experimental peaks matching this theoretical pattern, scales the theoretical distribution using
        :meth:`scale_theoretical_distribution`, and evaluates the quality of the fit using :attr:`scorer`.

        Parameters
        ----------
        peak : :class:`~.FittedPeak`
            The putative monoisotopic peak to use for interpolating an isotopic pattern
        error_tolerance : float
            Parts-per-million error tolerance for isotopic pattern matching
        charge : int
            The charge state to produce an isotopic pattern for
        charge_carrier : float, optional
            The charge carrier mass, defaults to |PROTON|

        Returns
        -------
        :class:`~.IsotopicFitRecord`
            The fitted isotopic pattern
        """
        tid = self.averagine.isotopic_cluster(
            peak.mz, charge, charge_carrier=charge_carrier,
            truncate_after=truncate_after, ignore_below=ignore_below)
        eid = self.match_theoretical_isotopic_distribution(
            tid, error_tolerance=error_tolerance)
        record = self._evaluate_theoretical_distribution(
            eid, tid, peak, charge)
        return record

    def _fit_peaks_at_charges(self, peak_charge_set, error_tolerance, charge_carrier=PROTON, truncate_after=0.8,
                              ignore_below=IGNORE_BELOW):
        """Given a set of candidate monoisotopic peaks and charge states, and a PPM error tolerance,
        fit each putative isotopic pattern.

        Calls :meth:`fit_theoretical_distribution` on each candidate.

        If a fit does not satisfy :attr:`scorer` `.reject`, it is discarded. If a fit has only one real peak
        and has a charge state greater than 1, it will also be discarded.

        Parameters
        ----------
        peak_charge_set : set
            The set of candidate (:class:`~.FittedPeak`, charge) tuples to try to fit
        error_tolerance : float
            Matching error tolerance
        charge_carrier : float, optional
            The charge carrier to use. Defaults to |PROTON|

        Returns
        -------
        set
            The set of :class:`~.IsotopicFitRecord` instances produced
        """
        results = []
        for peak, charge in peak_charge_set:
            if peak.mz < 1:
                continue
            fit = self.fit_theoretical_distribution(
                peak, error_tolerance, charge, charge_carrier, truncate_after,
                ignore_below)
            fit.missed_peaks = count_placeholders(fit.experimental)
            if not self._check_fit(fit):
                continue
            results.append(fit)
        return set(results)


try:
    from ms_deisotope._c.deconvoluter_base import AveragineDeconvoluterBase
except ImportError:
    pass


class AveragineDeconvoluter(AveragineDeconvoluterBase, ExhaustivePeakSearchDeconvoluterBase):
    """A Deconvoluter which uses an :title-reference:`averagine` [1] model to generate theoretical
    isotopic patterns for each peak to consider. Combines :class:`AveragineDeconvoluterBase` and
    :class:`ExhaustivePeakSearchDeconvoluterBase` to create a working Deconvoluter type.


    Attributes
    ----------
    averagine : :class:`~.AveragineCache`
        The averagine model and associated theoretical isotopic pattern cache to use
        to build theoretical isotopic patterns.
    peaklist : :class:`~.PeakSet`
        The collection of ms_peak_picker.FittedPeak instances and possible associated
        data to deconvolute.
    scorer : :class:`~.IsotopicFitterBase`
        The criterion for evaluating individual isotopic pattern fits
    verbose : bool
        How much diagnostic information to provide

    References
    ----------
    [1] Senko, M. W., Beu, S. C., & McLafferty, F. W. (1995). Determination of monoisotopic masses and ion populations
        for large biomolecules from resolved isotopic distributions. Journal of the American Society for Mass
        Spectrometry, 6(4), 229–233. http://doi.org/10.1016/1044-0305(95)00017-8
    """

    def __init__(self, peaklist, averagine=None, scorer=penalized_msdeconv,
                 use_subtraction=True, scale_method=SCALE_METHOD,
                 verbose=False, **kwargs):
        if averagine is None:
            averagine = AveragineCache(peptide, dict())
        else:
            if not isinstance(averagine, AveragineCache):
                averagine = AveragineCache(averagine, dict())
        self.peaklist = prepare_peaklist(peaklist)
        self.averagine = averagine
        self.scorer = scorer
        self._deconvoluted_peaks = []
        self.verbose = verbose

        super(AveragineDeconvoluter, self).__init__(
            use_subtraction, scale_method, merge_isobaric_peaks=True, **kwargs)


class MultiAveragineDeconvoluterBase(DeconvoluterBase):
    """A base class derived from :class:`DeconvoluterBase` which provides some common methods
    for fitting isotopic patterns using multiple Averagine models.
    """

    def _fit_peaks_at_charges(self, peak_charge_set, error_tolerance, charge_carrier=PROTON,
                              truncate_after=TRUNCATE_AFTER, ignore_below=IGNORE_BELOW):
        results = []
        for peak, charge in peak_charge_set:
            for averagine in self.averagines:
                if peak.mz < 1:
                    continue
                fit = self.fit_theoretical_distribution(
                    peak, error_tolerance, charge, averagine, charge_carrier=charge_carrier,
                    truncate_after=truncate_after, ignore_below=ignore_below)
                fit.missed_peaks = count_placeholders(fit.experimental)
                fit.data = averagine
                if not self._check_fit(fit):
                    continue
                results.append(fit)
        return set(results)


try:
    from ms_deisotope._c.deconvoluter_base import MultiAveragineDeconvoluterBase
except ImportError:
    pass


class MultiAveragineDeconvoluter(MultiAveragineDeconvoluterBase, ExhaustivePeakSearchDeconvoluterBase):
    """A Deconvoluter which uses multiple :title-reference:`averagine` [1] model to generate theoretical
    isotopic patterns for each peak to consider. Combines :class:`MultiAveragineDeconvoluterBase` and
    :class:`ExhaustivePeakSearchDeconvoluterBase` to create a working Deconvoluter type.

    This differs from :class:`AveragineDeconvoluter`, in that it will produce multiple isotopic fits for
    each (peak, charge) pair. This is advantageous when the isotopic patterns produced by different models
    are sufficiently different enough that they will favor different peak sets.

    Attributes
    ----------
    averagine : list of :class:`~.ms_deisotope.averagine.AveragineCache`
        The averagine models and associated theoretical isotopic pattern caches to use
        to build theoretical isotopic patterns.
    peaklist : :class:`~.ms_peak_picker.PeakSet`
        The collection of ms_peak_picker.FittedPeak instances and possible associated
        data to deconvolute.
    scorer : :class:`~.ms_deisotope.scoring.IsotopicFitterBase`
        The criterion for evaluating individual isotopic pattern fits
    verbose : bool
        How much diagnostic information to provide

    References
    ----------
    [1] Senko, M. W., Beu, S. C., & McLafferty, F. W. (1995). Determination of monoisotopic masses and ion populations
        for large biomolecules from resolved isotopic distributions. Journal of the American Society for Mass
        Spectrometry, 6(4), 229–233. http://doi.org/10.1016/1044-0305(95)00017-8
    """

    def __init__(self, peaklist, averagines=None, scorer=penalized_msdeconv,
                 use_subtraction=True, scale_method=SCALE_METHOD,
                 merge_isobaric_peaks=True, minimum_intensity=5.,
                 verbose=False, *args, **kwargs):
        self.peaklist = prepare_peaklist(peaklist)
        self.scorer = scorer
        self.use_subtraction = use_subtraction
        self.scale_method = scale_method

        cache_backend = dict
        if averagines is None:
            averagines = [peptide, glycopeptide, glycan]
        averagines = [
            AveragineCache(avg, backend=cache_backend()) if not isinstance(
                avg, AveragineCache) else avg
            for avg in averagines]
        self.averagines = averagines
        self.verbose = verbose

        self._deconvoluted_peaks = []

        super(MultiAveragineDeconvoluter, self).__init__(
            use_subtraction, scale_method, merge_isobaric_peaks,
            minimum_intensity, *args, **kwargs)


class AveraginePeakDependenceGraphDeconvoluter(AveragineDeconvoluter, PeakDependenceGraphDeconvoluterBase):
    """A Deconvoluter which uses an :title-reference:`averagine` [1] model to generate theoretical
    isotopic patterns for each peak to consider, using a peak dependence graph to solve complex mass
    spectra.

    Extends :class:`AveragineDeconvoluter` to include features from
    :class:`PeakDependenceGraphDeconvoluterBase` making it suitable for deconvoluting complex spectra where
    peak overlaps are common.

    Attributes
    ----------
    peaklist : :class:`~.PeakSet`
        The centroided mass spectrum to deconvolute
    scorer : :class:`~.IsotopicFitterBase`
        The criterion for evaluating individual isotopic pattern fits
    averagine : :class:`~.AveragineCache`
        The averagine model and associated theoretical isotopic pattern cache to use
        to build theoretical isotopic patterns.
    max_missed_peaks : int
        The maximum number of missing peaks to tolerate in an isotopic fit
    peak_dependency_network : :class:`~.PeakDependenceGraph`
        The peak dependence graph onto which isotopic fit dependences on peaks
        are constructed and solved.
    merge_isobaric_peaks : bool
        If multiple passes produce peaks with identical mass values,
        should those peaks be summed
    minimum_intensity : float
        Experimental peaks whose intensity is below this level will be ignored
        by peak querying methods
    scale_method : str
        The name of the method to use to scale theoretical isotopic pattern intensities
        to match the experimental isotopic pattern. For a description of options, see
        :meth:`~.TheoreticalIsotopicPattern.scale`.
    use_subtraction : bool
        Whether or not to apply a subtraction procedure to experimental peaks after they
        have been fitted. This is only necessary if the same signal may be examined multiple
        times as in a multi-pass method or when peak dependence is not considered
    verbose : bool
        Produce extra logging information

    References
    ----------
    [1] Senko, M. W., Beu, S. C., & McLafferty, F. W. (1995). Determination of monoisotopic masses and ion populations
        for large biomolecules from resolved isotopic distributions. Journal of the American Society for Mass
        Spectrometry, 6(4), 229–233. http://doi.org/10.1016/1044-0305(95)00017-8
    """

    def __init__(self, peaklist, *args, **kwargs):
        super(AveraginePeakDependenceGraphDeconvoluter,
              self).__init__(peaklist, *args, **kwargs)


class MultiAveraginePeakDependenceGraphDeconvoluter(MultiAveragineDeconvoluter, PeakDependenceGraphDeconvoluterBase):
    """Extends :class:`MultiAveragineDeconvoluter` to include features from
    :class:`PeakDependenceGraphDeconvoluterBase` making it suitable for deconvoluting complex spectra where
    peak overlaps are common.

    Attributes
    ----------
    peaklist : :class:`~.ms_peak_picker.PeakSet`
        The centroided mass spectrum to deconvolute
    scorer : :class:`~.IsotopicFitterBase`
        The criterion for evaluating individual isotopic pattern fits
    averagine : list of :class:`~.ms_deisotope.averagine.AveragineCache`
        The averagine model and associated theoretical isotopic pattern cache to use
        to build theoretical isotopic patterns.
    max_missed_peaks : int
        The maximum number of missing peaks to tolerate in an isotopic fit
    peak_dependency_network : :class:`~.PeakDependenceGraph`
        The peak dependence graph onto which isotopic fit dependences on peaks
        are constructed and solved.
    merge_isobaric_peaks : bool
        If multiple passes produce peaks with identical mass values,
        should those peaks be summed
    minimum_intensity : float
        Experimental peaks whose intensity is below this level will be ignored
        by peak querying methods
    scale_method : str
        The name of the method to use to scale theoretical isotopic pattern intensities
        to match the experimental isotopic pattern. For a description of options, see
        :meth:`~.TheoreticalIsotopicPattern.scale`.
    use_subtraction : bool
        Whether or not to apply a subtraction procedure to experimental peaks after they
        have been fitted. This is only necessary if the same signal may be examined multiple
        times as in a multi-pass method or when peak dependence is not considered
    verbose : bool
        Produce extra logging information

    """

    def __init__(self, peaklist, *args, **kwargs):
        super(MultiAveraginePeakDependenceGraphDeconvoluter,
              self).__init__(peaklist, *args, **kwargs)



from ms_deisotope._c.feature_map.lcms_feature cimport LCMSFeature
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM


cdef class LCMSFeatureMap(object):
    cdef:
        public list features

    cpdef list _find_all(self, double mz, double error_tolerance)
    cpdef LCMSFeature _search(self, double mz, double error_tolerance)

cpdef tuple binary_search_with_flag(list array, double mz, double error_tolerance)
cdef long binary_search(list array, double mz, double error_tolerance)
cdef void search_sweep(list array, double mz, double error_tolerance, long* loout, long* hiout)




import unittest

from ms_deisotope.data_source import MzXMLLoader
from ms_deisotope.test.common import datafile
from ms_deisotope.data_source import infer_type


class TestMzXMLLoaderScanBehavior(unittest.TestCase):
    path = datafile("microscans.mzXML")

    @property
    def reader(self):
        return infer_type.MSFileLoader(self.path)

    @property
    def first_scan(self):
        return self.reader.next().precursor

    def test_id(self):
        loader = self.reader
        scan = next(loader).precursor
        self.assertEqual(scan.id, "210")
        scan = loader.get_scan_by_id("210")
        self.assertEqual(scan.id, "210")

    def test_start_from_scan(self):
        loader = self.reader
        time = 0.4856916666666667
        bunch = next(loader.start_from_scan(rt=time))
        self.assertAlmostEqual(bunch.precursor.scan_time, time, 3)
        ix = bunch.precursor.index
        assert next(loader.start_from_scan(index=ix)).precursor == bunch.precursor

    def test_polarity(self):
        self.assertEqual(self.first_scan.polarity, 1)

    def test_index(self):
        self.assertEqual(self.first_scan.polarity, 1)

    def test_arrays(self):
        self.assertEqual(len(self.first_scan.arrays), 2)

    def test_precursor_info(self):
        self.assertEqual(self.first_scan.precursor_information, None)

    def test_file_description(self):
        file_info = self.reader.file_description()
        source_file = file_info.source_files[0]
        assert source_file.name == "AGP_tryptic_300ng_3microscans_glycoproteomics_nCE_27-35.raw"
        assert "location" not in source_file.parameters

    def test_data_processing(self):
        proc_info = self.reader.data_processing()
        assert len(proc_info) == 2


if __name__ == '__main__':
    unittest.main()



# cython: embedsignature=True

cimport cython
from libc.math cimport fabs, sqrt, log, ceil, floor
from libc.stdlib cimport malloc, free
import operator

from cpython.list cimport PyList_New, PyList_GetItem, PyList_Size, PyList_GET_ITEM, PyList_SET_ITEM, PyList_GET_SIZE
from cpython.sequence cimport PySequence_List

from brainpy._c.isotopic_distribution cimport TheoreticalPeak
from ms_peak_picker._c.peak_set cimport PeakSet, FittedPeak


@cython.nonecheck(False)
@cython.cdivision(True)
cdef bint isclose(double x, double y, double rtol=1.e-5, double atol=1.e-8):
    return abs(x-y) <= (atol + rtol * abs(y))


@cython.freelist(50000)
cdef class IsotopicFitRecord(object):
    """Describes a single isotopic pattern fit, comparing how well an
    experimentally observed sequence of peaks matches a theoretical isotopic
    pattern.
    
    IsotopicFitRecord instances are hashable and orderable (by score).

    Attributes
    ----------
    charge : int
        The charge state used to generate the theoretical pattern
    data : object
        An arbitrary Python object containing extra information
    experimental : list of FittedPeak
        The observed experimental peaks to be fitted
    missed_peaks : int
        The number of peaks in the theoretical pattern that do not
        have a matching experimental peak
    monoisotopic_peak : FittedPeak
        The fitted peak which corresponds to the monoisotopic peak
    score : float
        The score assigned to the fit by an IsotopicFitter object
    seed_peak : FittedPeak
        The peak that was used to initiate the fit. This may be unused
        if not using an Averagine method
    theoretical : TheoreticalIsotopicPattern
        The theoretical isotopic pattern being fitted on the experimental data
    """
    def __init__(self, FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical, list experimental,
                  object data=None, int missed_peaks=0):
        self.seed_peak = seed_peak
        self.score = score
        self.charge = charge
        self.experimental = experimental
        self.theoretical = theoretical
        self.monoisotopic_peak = experimental[0]
        self.data = data
        self.missed_peaks = missed_peaks
        self._hash = -1

    @staticmethod
    cdef IsotopicFitRecord _create(FittedPeak seed_peak, double score, int charge, TheoreticalIsotopicPattern theoretical,
                                   list experimental, object data=None, int missed_peaks=0):
        cdef IsotopicFitRecord inst
        inst = IsotopicFitRecord.__new__(IsotopicFitRecord)
        inst.seed_peak = seed_peak
        inst.score = score
        inst.charge = charge
        inst.theoretical = theoretical
        inst.experimental = experimental
        inst.monoisotopic_peak = <FittedPeak>PyList_GetItem(experimental, 0)
        inst.data = data
        inst.missed_peaks = missed_peaks
        inst._hash = -1
        return inst

    def clone(self):
        return IsotopicFitRecord._create(self.seed_peak, self.score, self.charge, self.theoretical,
                              self.experimental, self.data, self.missed_peaks)

    def __reduce__(self):
        return self.__class__, (self.seed_peak, self.score, self.charge, self.theoretical,
                                self.experimental, self.data, self.missed_peaks)

    cpdef bint _eq(self, IsotopicFitRecord other):
        cdef:
            bint val
            size_t i, n
            FittedPeak fp1, fp2
            TheoreticalPeak tp1, tp2

        val = self.score == other.score
        if not val:
            return val
        val = self.charge == other.charge
        if not val:
            return val
        n = PyList_GET_SIZE(self.experimental)
        val = n == PyList_GET_SIZE(other.experimental)
        if not val:
            return val
        for i in range(n):
            fp1 = <FittedPeak>PyList_GET_ITEM(self.experimental, i)
            fp2 = <FittedPeak>PyList_GET_ITEM(other.experimental, i)
            val = fp1._eq(fp2)
            if not val:
                return val
        val = self.theoretical._eq_inst(other.theoretical)
        if not val:
            return val
        if self.data is not None or other.data is not None:
            val = val & (self.data == other.data)
        return val

    cpdef bint _ne(self, IsotopicFitRecord other):
        return not (self._eq(other))

    cpdef bint _lt(self, IsotopicFitRecord other):
        return self.score < other.score

    cpdef bint _gt(self, IsotopicFitRecord other):
        return self.score > other.score

    def __richcmp__(self, IsotopicFitRecord other, int code):
        if other is None:
            if code == 3:
                return True
            else:
                return False

        if code == 0:
            return self._lt(other)
        elif code == 2:
            return self._eq(other)
        elif code == 3:
            return not (self._eq(other))
        elif code == 4:
            return self._gt(other)


    def __hash__(self):
        if self._hash == -1:
            self._hash = hash((self.monoisotopic_peak.mz, self.charge))
        return self._hash

    def __iter__(self):
        yield self.score
        yield self.charge
        yield self.experimental
        yield self.theoretical

    @property
    def npeaks(self):
        return len(self.experimental)

    def __repr__(self):
        return "IsotopicFitRecord(score=%0.5f, charge=%d, npeaks=%d, monoisotopic_mz=%0.5f)" % (
            self.score, self.charge, self.npeaks, self.monoisotopic_peak.mz)


cdef class FitSelectorBase(object):
    """An object that controls the filtering and
    selection of IsotopicFitRecord

    Attributes
    ----------
    minimum_score : int
        The minimum score needed to be a candidate for selection. If the
        FitSelector is `minimizing` it is the maximal score to be a candidate.
    """
    def __init__(self, minimum_score=0):
        self.minimum_score = minimum_score

    def __reduce__(self):
        return self.__class__, (self.minimum_score,)

    cpdef IsotopicFitRecord best(self, object results):
        raise NotImplementedError()

    cdef IsotopicFitRecord _best_from_set(self, set results):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        return self.best(*args, **kwargs)

    cpdef bint reject(self, IsotopicFitRecord result):
        raise NotImplementedError()

    cpdef bint reject_score(self, double score):
        raise NotImplementedError()

    cpdef bint is_maximizing(self):
        return False

    def __repr__(self):
        return "{self.__class__.__name__}(minimum_score={self.minimum_score})".format(self=self)


cdef class MinimizeFitSelector(FitSelectorBase):
    """A FitSelector which tries to minimize the score of the best fit.
    """
    cpdef IsotopicFitRecord best(self, object results):
        """Returns the IsotopicFitRecord with the smallest score
        
        Parameters
        ----------
        results : list of IsotopicFitRecord
            List of isotopic fits to select the most optimal case from
        
        Returns
        -------
        IsotopicFitRecord
            The most optimal fit
        """
        if isinstance(results, set):
            return self._best_from_set(<set>results)
        return min(results, key=operator.attrgetter("score"))

    cdef IsotopicFitRecord _best_from_set(self, set results):
        cdef:
            IsotopicFitRecord fit, best_fit
            object obj
            double best_score

        best_fit = None
        best_score = INFINITY
        for obj in results:
            fit = <IsotopicFitRecord>obj
            if fit.score < best_score:
                best_score = fit.score
                best_fit = fit
        return best_fit


    cpdef bint reject(self, IsotopicFitRecord fit):
        """Decide whether the fit should be discarded for having too
        large a score. Compares against :attr:`minimum_score`
        
        Parameters
        ----------
        fit : IsotopicFitRecord
        
        Returns
        -------
        bool
        """
        return fit.score > self.minimum_score

    cpdef bint reject_score(self, double score):
        return score > self.minimum_score

    cpdef bint is_maximizing(self):
        """Returns that this is *not* a maximizing selector

        Returns
        -------
        False
        """
        return False


cdef class MaximizeFitSelector(FitSelectorBase):
    """A FitSelector which tries to maximize the score of the best fit.
    """
    cpdef IsotopicFitRecord best(self, object results):
        """Returns the IsotopicFitRecord with the largest score
        
        Parameters
        ----------
        results : list of IsotopicFitRecord
            List of isotopic fits to select the most optimal case from
        
        Returns
        -------
        IsotopicFitRecord
            The most optimal fit
        """
        if isinstance(results, set):
            return self._best_from_set(<set>results)
        return max(results, key=operator.attrgetter("score"))

    cdef IsotopicFitRecord _best_from_set(self, set results):
        cdef:
            IsotopicFitRecord fit, best_fit
            double best_score

        best_fit = None
        best_score = -INFINITY
        for obj in results:
            fit = <IsotopicFitRecord>obj
            if fit.score > best_score:
                best_score = fit.score
                best_fit = fit
        return best_fit

    cpdef bint reject(self, IsotopicFitRecord fit):
        """Decide whether the fit should be discarded for having too
        small a score. Compares against :attr:`minimum_score`
        
        Parameters
        ----------
        fit : IsotopicFitRecord
        
        Returns
        -------
        bool
        """
        return fit.score < self.minimum_score

    cpdef bint reject_score(self, double score):
        return score < self.minimum_score

    cpdef bint is_maximizing(self):
        """Returns that this *is* a maximizing selector

        Returns
        -------
        False
        """
        return True


cdef class IsotopicFitterBase(object):
    '''A base class for Isotopic Pattern Fitters, objects
    which given a set of experimental peaks and a set of matching
    theoretical peaks, returns a fit score describing how good the
    match is.

    An IsotopicFitter may be optimal when the score is small (minimizing)
    or when the score is large (maximizing), and the appropriate
    :class:`FitSelectorBase` type will be used for :attr:`select`. This
    will also be reflected by :meth:`is_maximizing`.
    '''

    def __init__(self, score_threshold=0.5):
        self.select = MinimizeFitSelector(score_threshold)

    def __reduce__(self):
        return self.__class__, (0,), self.__getstate__()

    def __getstate__(self):
        return (self.select,)

    def __setstate__(self, state):
        self.select, = state

    def evaluate(self, PeakSet peaklist, list observed, list expected, **kwargs):
        """Evaluate a pair of peak lists for goodness-of-fit.

        Parameters
        ----------
        peaklist : :class:`~.PeakSet`
            The full set of all experimental peaks
        observed : list
            The list of experimental peaks that are part of this fit
        expected : list
            The list of theoretical peaks that are part of this fit
        **kwargs

        Returns
        -------
        float
            The score
        """
        return self._evaluate(peaklist, observed, expected)

    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        """Invokes :meth:`evaluate`
        
        Parameters
        ----------
        *args
            Forwarded to :meth:`evaluate`
        **kwargs
            Forwarded to :meth:`evaluate`
        
        Returns
        -------
        float
            The score
        """
        return self.evaluate(*args, **kwargs)

    cpdef bint reject(self, IsotopicFitRecord fit):
        """Test whether this fit is too poor to be used

        Parameters
        ----------
        fit : :class:`~IsotopicFitRecord`
            The fit to test

        Returns
        -------
        bool
        """
        return self.select.reject(fit)

    cpdef bint reject_score(self, double score):
        return self.select.reject_score(score)

    cpdef bint is_maximizing(self):
        """Whether or not this fitter's score gets better as it grows

        Returns
        -------
        bool
            Whether or not this fitter is a maximizing fitter
        """
        return self.select.is_maximizing()

    cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):
        return self

    def configure(self, DeconvoluterBase deconvoluter, **kwargs):
        return self._configure(deconvoluter, kwargs)

    def __repr__(self):
        return "{self.__class__.__name__}({fields})".format(self=self, fields=self.__getstate__())


cdef double sum_intensity_theoretical(list peaklist):
    cdef:
        double summed
        size_t i
        TheoreticalPeak peak
    summed = 0
    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
        summed += peak.intensity
    return summed    


cdef double sum_intensity_fitted(list peaklist):
    cdef:
        double summed
        size_t i
        FittedPeak peak
    summed = 0
    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
        summed += peak.intensity
    return summed


@cython.cdivision
cdef double* normalize_intensity_theoretical(list peaklist, double* out, double total):
    cdef:
        size_t i
        TheoreticalPeak peak

    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
        out[i] = peak.intensity / total


@cython.cdivision
cdef double* normalize_intensity_fitted(list peaklist, double* out, double total):
    cdef:
        size_t i
        FittedPeak peak

    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
        out[i] = peak.intensity / total


cdef class GTestFitter(IsotopicFitterBase):
    r"""Evaluate an isotopic fit using a G-test

    .. math::
        G = 2\sum_i^n{o_i * ({log}o_i - {log}e_i)}

    where :math:`o_i` is the intensity of the ith experimental peak
    and :math:`e_i` is the intensity of the ith theoretical peak.
    """
    @cython.cdivision
    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            double g_score, obs, theo, log_ratio
            size_t n

        n = PyList_GET_SIZE(observed)

        g_score = 0.
        for i in range(n):
            obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity
            theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity

            log_ratio = log(obs) - log(theo)
            g_score += obs * log_ratio

        return g_score * 2.


cdef class ScaledGTestFitter(IsotopicFitterBase):
    r"""Evaluate an isotopic fit using a G-test after normalizing the
    list of experimental and theoretical peaks to both sum to 1.

    .. math::
        G = 2\sum_i^n{o_i * ({log}o_i - {log}e_i)}

    where :math:`o_i` is the intensity of the ith experimental peak
    and :math:`e_i` is the intensity of the ith theoretical peak.
    """

    @cython.cdivision
    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            double total_observed
            double total_expected
            double g_score, obs, theo, log_ratio
            size_t n
        total_observed = sum_intensity_fitted(observed)
        total_expected = sum_intensity_theoretical(expected)
        n = PyList_GET_SIZE(observed)

        g_score = 0.
        for i in range(n):
            obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_observed
            theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_expected

            log_ratio = log(obs) - log(theo)
            g_score += obs * log_ratio

        return g_score * 2.


cdef ScaledGTestFitter g_test_scaled

g_test_scaled = ScaledGTestFitter()


cdef double max_intensity_theoretical(list peaklist):
    cdef:
        double maximum
        size_t i
        TheoreticalPeak peak
    maximum = 0
    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <TheoreticalPeak>PyList_GET_ITEM(peaklist, i)
        if peak.intensity > maximum:
            maximum = peak.intensity
    return maximum


cdef double max_intensity_fitted(list peaklist):
    cdef:
        double maximum
        size_t i
        FittedPeak peak
    maximum = 0
    for i in range(PyList_GET_SIZE(peaklist)):
        peak = <FittedPeak>PyList_GET_ITEM(peaklist, i)
        if peak.intensity > maximum:
            maximum = peak.intensity
    return maximum


cdef class LeastSquaresFitter(IsotopicFitterBase):
    r"""Evaluate an isotopic fit using least squares coefficient of
    determination :math:`R^2`.

    .. math::
        {\hat e_i} &= e_i / max(e)

        {\hat t_i} &= t_i / max(t)

        {\hat t} &= \sum_i^n {\hat t_i}^2

        R^2 &= \frac{1}{{\hat t}}\sum_i^n ({\hat e_i} - {\hat t_i})^2

    where :math:`e_i` is the ith experimental peak intensity and :math:`t_i`
    is the ith theoretical peak intensity
    """
    @cython.cdivision
    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            double exp_max, theo_max, sum_of_squared_errors, sum_of_squared_theoreticals
            double normed_theo, normed_expr
            size_t i
            TheoreticalPeak t
            FittedPeak e

        exp_max = max_intensity_fitted(observed)
        theo_max = max_intensity_theoretical(expected)

        sum_of_squared_errors = 0
        sum_of_squared_theoreticals = 0

        assert len(observed) == len(expected)

        # for e, t in zip(observed, expected):
        for i in range(PyList_GET_SIZE(observed)):
            e = <FittedPeak>PyList_GetItem(observed, i)
            t = <TheoreticalPeak>PyList_GetItem(expected, i)
            normed_expr = e.intensity / exp_max
            normed_theo = t.intensity / theo_max
            sum_of_squared_errors += (normed_theo - normed_expr) ** 2
            sum_of_squared_theoreticals += normed_theo ** 2
        return sum_of_squared_errors / sum_of_squared_theoreticals


cdef LeastSquaresFitter least_squares

least_squares = LeastSquaresFitter()


cdef class ChiSquareFitter(IsotopicFitterBase):

    @cython.cdivision
    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            double running_total
            size_t i
            TheoreticalPeak t
            FittedPeak e
        running_total = 0
        for i in range(PyList_GET_SIZE(observed)):
            e = <FittedPeak>PyList_GetItem(observed, i)
            t = <TheoreticalPeak>PyList_GetItem(expected, i)
            running_total += ((e - t) ** 2) / t
        return running_total


@cython.cdivision
cdef double ms_deconv_score_peak(FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=0.02, double minimum_signal_to_noise=1) nogil:
    cdef:
        double mass_error, abundance_diff
    if obs.signal_to_noise < minimum_signal_to_noise:
        return 0.

    mass_error = fabs(obs.mz - theo.mz)

    if mass_error <= mass_error_tolerance:
        mass_accuracy = 1 - mass_error / mass_error_tolerance
    else:
        return 0.

    if obs.intensity < theo.intensity and (((theo.intensity - obs.intensity) / obs.intensity) <= 1):
        abundance_diff = 1 - ((theo.intensity - obs.intensity) / obs.intensity)
    elif obs.intensity >= theo.intensity and (((obs.intensity - theo.intensity) / obs.intensity) <= 1):
        abundance_diff = sqrt(1 - ((obs.intensity - theo.intensity) / obs.intensity))
    else:
        return 0.
    score = sqrt(theo.intensity) * mass_accuracy * abundance_diff
    return score


cdef class MSDeconVFitter(IsotopicFitterBase):
    '''An implementation of the scoring function used in :title-reference:`MSDeconV`

    References
    ----------
    Liu, X., Inbar, Y., Dorrestein, P. C., Wynne, C., Edwards, N., Souda, P., …
    Pevzner, P. A. (2010). Deconvolution and database search of complex tandem
    mass spectra of intact proteins: a combinatorial approach. Molecular & Cellular
    Proteomics : MCP, 9(12), 2772–2782. https://doi.org/10.1074/mcp.M110.002766
    '''
    def __init__(self, minimum_score=10, mass_error_tolerance=0.02):
        self.mass_error_tolerance = mass_error_tolerance
        self.select = MaximizeFitSelector()
        self.select.minimum_score = minimum_score

    @cython.cdivision
    cdef double score_peak(self, FittedPeak obs, TheoreticalPeak theo, double mass_error_tolerance=0.02, double minimum_signal_to_noise=1) nogil:
        return ms_deconv_score_peak(obs, theo, mass_error_tolerance, minimum_signal_to_noise)

    def evaluate(self, PeakSet peaklist, list observed, list expected):
        return self._evaluate(peaklist, observed, expected)

    def __reduce__(self):
        return self.__class__, (0,), self.__getstate__()

    def __getstate__(self):
        return (self.select, self.mass_error_tolerance)

    def __setstate__(self, state):
        self.select, self.mass_error_tolerance = state

    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            size_t i, n
            FittedPeak obs
            TheoreticalPeak theo
            double score

        n = PyList_GET_SIZE(observed)
        score = 0
        for i in range(n):
            obs = <FittedPeak>PyList_GET_ITEM(observed, i)
            theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
            score += self.score_peak(obs, theo, self.mass_error_tolerance, 1)

        return score


cdef class PenalizedMSDeconVFitter(IsotopicFitterBase):
    r'''An Isotopic Fitter which uses the :class:`MSDeconVFitter` score
    weighted by 1 - :attr:`penalty_factor` * :class:`ScaledGTestFitter` score

    .. math::
        S(e, t) = M(e, t) * (1 - G(e, t))

    where :math:`e` is the experimental peak list and :math:`t` is the theoretical
    peak list
    '''
    def __init__(self, minimum_score=10, penalty_factor=1, mass_error_tolerance=0.02):
        self.select = MaximizeFitSelector(minimum_score)
        self.penalty_factor = penalty_factor
        self.mass_error_tolerance = mass_error_tolerance

    def __reduce__(self):
        return self.__class__, (0,), self.__getstate__()

    def __getstate__(self):
        return (self.select, self.penalty_factor, self.mass_error_tolerance)

    def __setstate__(self, state):
        self.select, self.penalty_factor, self.mass_error_tolerance = state

    def evaluate(self, PeakSet peaklist, list observed, list expected):
        return self._evaluate(peaklist, observed, expected)

    @cython.cdivision
    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        cdef:
            size_t i, n
            FittedPeak obs
            TheoreticalPeak theo
            double score, total_intensity_observed, total_intensity_expected
            double penalty, _obs, _theo, log_ratio

        total_intensity_observed = 0
        total_intensity_expected = 0

        n = PyList_GET_SIZE(observed)
        score = 0
        for i in range(n):
            obs = <FittedPeak>PyList_GET_ITEM(observed, i)
            theo = <TheoreticalPeak>PyList_GET_ITEM(expected, i)
            score += ms_deconv_score_peak(obs, theo, self.mass_error_tolerance, 1)
            total_intensity_observed += obs.intensity
            total_intensity_expected += theo.intensity
        penalty = 0
        for i in range(n):
            _obs = (<FittedPeak>PyList_GET_ITEM(observed, i)).intensity / total_intensity_observed
            _theo = (<TheoreticalPeak>PyList_GET_ITEM(expected, i)).intensity / total_intensity_expected

            log_ratio = log(_obs) - log(_theo)
            penalty += _obs * log_ratio
        penalty = abs(2 * penalty)
        score *= ((1 - penalty * self.penalty_factor))
        return score


cdef class FunctionScorer(IsotopicFitterBase):

    def __init__(self, function, minimum_score=10, selector_type=MaximizeFitSelector):
        self.function = function
        self.select = selector_type(minimum_score)

    def __getstate__(self):
        return self.select, self.function

    def __setstate__(self, state):
        self.select, self.function = state

    def __reduce__(self):
        return self.__class__, (None,), self.__getstate__()

    cpdef double _evaluate(self, PeakSet peaklist, list observed, list expected):
        return self.function(observed, expected)


cdef class InterferenceDetection(object):
    def __init__(self, PeakSet peaklist):
        self.peaklist = peaklist

    def __reduce__(self):
        return self.__class__, (None,), self.__getstate__()

    def __getstate__(self):
        return (self.peaklist,)

    def __setstate__(self, state):
        self.peaklist, = state

    cdef double detect_interference(self, list experimental_peaks, double lower=-1, double upper=-1):
        cdef:
            FittedPeak min_peak, max_peak
            PeakSet region
            double included_intensity, region_intensity, score
            size_t n

        n = PyList_GET_SIZE(experimental_peaks)
        if n == 0:
            return 0.0

        min_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, 0)
        max_peak = <FittedPeak>PyList_GET_ITEM(experimental_peaks, PyList_GET_SIZE(experimental_peaks) - 1)

        if lower < 0:
            lower = min_peak.mz - min_peak.full_width_at_half_max
        if upper < 0:
            upper = max_peak.mz + max_peak.full_width_at_half_max

        region = self.peaklist._between(
            lower,
            upper)

        included_intensity = sum_intensity_fitted(experimental_peaks)
        region_intensity = sum_intensity_fitted(list(region))
        if region_intensity == 0:
            return 0.0

        score = 1 - (included_intensity / region_intensity)
        return score

    def __call__(self, list experimental_peaks, lower=None, upper=None):
        if lower is None:
            lower = -1
        if upper is None:
            upper = -1
        return self.detect_interference(experimental_peaks, lower, upper)

cdef class DistinctPatternFitter(IsotopicFitterBase):

    def __init__(self, minimum_score=0.3, peak_count_scale=1.5, domain_scale=100.):
        self.select = MinimizeFitSelector(minimum_score)
        self.interference_detector = None
        self.g_test_scaled = ScaledGTestFitter()
        self.peak_count_scale = peak_count_scale
        self.domain_scale = domain_scale

    def __getstate__(self):
        return self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale

    def __setstate__(self, state):
        self.select, self.interference_detector, self.g_test_scaled, self.peak_count_scale, self.domain_scale = state

    cpdef IsotopicFitterBase _configure(self, DeconvoluterBase deconvoluter, dict kwargs):
        self.interference_detector = InterferenceDetection(deconvoluter.peaklist)
        return self    

    cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):
        cdef:
            double score
            double npeaks

        npeaks = PyList_GET_SIZE(experimental)

        if self.interference_detector is None:
            self.interference_detector = InterferenceDetection(peaklist)

        score = self.g_test_scaled._evaluate(peaklist, experimental, theoretical)
        score *= abs((self.interference_detector.detect_interference(experimental) + 0.00001) / (
            npeaks * self.peak_count_scale)) * self.domain_scale
        return score


cdef double percentile(double[:] N, double percent):
    cdef:
        double k, f, c, d0, d1
    k = (N.shape[0] - 1) * percent
    f = floor(k)
    c = ceil(k)
    if f == c:
        return N[<size_t>(k)]
    d0 = N[<size_t>(f)] * (c - k)
    d1 = N[<size_t>(c)] * (k - f)
    return d0 + d1


cdef class ScaledPenalizedMSDeconvFitter(IsotopicFitterBase):

    def __init__(self,  minimum_score=0.3, penalty_factor=1., mass_error_tolerance=0.02):
        self.select = MaximizeFitSelector(minimum_score)
        self.scale_factor = 0.
        self.scorer = PenalizedMSDeconVFitter(
            penalty_factor=penalty_factor, mass_error_tolerance=mass_error_tolerance)

    def __getstate__(self):
        return self.select, self.scale_factor, self.scorer

    def __setstate__(self, state):
        self.select, self.scale_factor, self.scorer = state

    @property
    def penalty_factor(self):
        return self.scorer.penalty_factor

    @penalty_factor.setter
    def penalty_factor(self, value):
        self.scorer.penalty_factor = value

    cpdef double _calculate_scale_factor(self, PeakSet peaklist):
        cdef:
            size_t i
            FittedPeak peak
            double maximum
            double intensity
            PeakSet peaks

        maximum = 0.
        peaks = peaklist.peaks
        for i in range(len(peaks)):
            peak = peaks.getitem(i)
            if peak.intensity > maximum:
                maximum = peak.intensity
        return maximum

    cdef void scale_fitted_peaks(self, list experimental, double factor):
        cdef:
            size_t i
            FittedPeak peak
        for i in range(len(experimental)):
            peak = <FittedPeak>PyList_GET_ITEM(experimental, i)
            peak.intensity *= factor        

    cdef void scale_theoretical_peaks(self, list theoretical, double factor):
        cdef:
            size_t i
            TheoreticalPeak peak
        for i in range(len(theoretical)):
            peak = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)
            peak.intensity *= factor

    def evaluate(self, PeakSet peaklist, list observed, list expected):
        return self._evaluate(peaklist, observed, expected)

    cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):
        cdef:
            double score
        if self.scale_factor < 1:
            self.scale_factor = self._calculate_scale_factor(peaklist)
        self.scale_fitted_peaks(experimental, 1. / self.scale_factor)
        self.scale_theoretical_peaks(theoretical, 1. / self.scale_factor)
        score = self.scorer._evaluate(peaklist, experimental, theoretical)
        self.scale_fitted_peaks(experimental, self.scale_factor)
        self.scale_theoretical_peaks(theoretical, self.scale_factor)
        return score


cdef class DotProductFitter(IsotopicFitterBase):

    def __init__(self,  minimum_score=100):
        self.select = MaximizeFitSelector(minimum_score)

    cpdef double _evaluate(self, PeakSet peaklist, list experimental, list theoretical):
        cdef:
            size_t i, n
            FittedPeak obs
            TheoreticalPeak theo
            double score
        n = PyList_GET_SIZE(experimental)
        score = 0
        for i in range(n):
            obs = <FittedPeak>PyList_GET_ITEM(experimental, i)
            theo = <TheoreticalPeak>PyList_GET_ITEM(theoretical, i)
            score += obs.intensity * theo.intensity
        return score



import unittest
import io

from click.testing import CliRunner

from ms_deisotope.data_source.mzml import MzMLLoader
from ms_deisotope.tools import indexing, conversion
from ms_deisotope.test.common import datafile


def test_describe():
    runner = CliRunner()

    path = datafile("small.mzML")
    result = runner.invoke(indexing.describe, [path])
    lines = result.output.splitlines()
    assert "small.mzML" in lines[0]
    assert lines[1] == "File Format: mzML format"
    assert lines[2] == "ID Format: Thermo nativeID format"
    assert lines[3] == "Format Supports Random Access: True"


def test_mgf():
    runner = CliRunner()

    path = datafile("small.mzML")
    result = runner.invoke(conversion.mgf, [path, '-'])
    lines = result.output.splitlines()
    count = 0
    for line in lines:
        if "BEGIN" in line:
            count += 1
    assert count == 34


def test_mzml():
    runner = CliRunner()

    path = datafile("small.mzML")
    result = runner.invoke(conversion.mzml, ['-p', '-c', path, '-'])
    buff = io.BytesIO(result.output.encode("utf-8"))
    reader = MzMLLoader(buff)
    n = len(reader)
    assert n == 48



from collections import deque


def bisect_at_score(scored_dependencies, splitting_value):
    lo = 0
    hi = len(scored_dependencies)

    while hi != lo:
        mid = (hi + lo) / 2
        mid_fit = scored_dependencies[mid]
        if hi == lo + 1:
            return mid
        elif abs(mid_fit.score - splitting_value) < 1e-3:
            return mid
        elif mid_fit.score > splitting_value:
            lo = mid
        elif mid_fit.score < splitting_value:
            hi = mid


def bisect_at_start(clusters, mz):
    lo = 0
    hi = len(clusters)
    while hi != lo:
        mid = (hi + lo) / 2
        mid_cluster = clusters[mid]
        if hi == lo + 1:
            return mid
        elif abs(mid_cluster.start - mz) < 1e-3:
            return mid
        elif mid_cluster.start < mz:
            hi = mid
        elif mid_cluster.start > mz:
            lo = mid


class GeneratorQueue(object):  # pragma: no cover
    sentinel = object()

    def __init__(self, queue=None):
        if queue is None:
            queue = deque()
        self.queue = queue
        self.active = None
        self._peek = self.sentinel

    def extend(self, iterable):
        self.queue.appendleft(iter(iterable))

    def load_next_generator(self):
        self.active = self.queue.popleft()

    def get_next_value(self):
        if self._peek is not self.sentinel:
            value = self._peek
            self._peek = self.sentinel
        else:
            try:
                value = next(self.active)
            except (StopIteration, TypeError):
                try:
                    self.load_next_generator()
                    value = self.get_next_value()
                except IndexError:
                    print("Terminating with size", len(self.queue))
                    raise StopIteration()
        return value

    def __next__(self):
        value = self.get_next_value()
        return value

    def next(self):
        return self.__next__()

    def __iter__(self):
        while True:
            yield self.next()
        print("Terminating with size", len(self.queue))

    def __len__(self):
        return len(self.queue)

    @property
    def peek(self):
        if self._peek is self.sentinel:
            try:
                self._peek = self.get_next_value()
            except StopIteration:
                self._peek = self.sentinel
        return self._peek

    def hasnext(self):
        return self.peek is not self.sentinel
